from __future__ import annotations

import sys
from dataclasses import replace
from datetime import datetime, time
from decimal import Decimal
from typing import Any
from unittest import mock

import pytest
from yarl import URL

from neuro_config_client.entities import (
    AMDGPU,
    ACMEEnvironment,
    AddNodePoolRequest,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    AWSCloudProvider,
    AWSCredentials,
    AWSStorage,
    AzureCloudProvider,
    AzureCredentials,
    AzureReplicationType,
    AzureStorage,
    AzureStorageTier,
    BucketsConfig,
    CloudProviderOptions,
    CloudProviderType,
    ClusterLocationType,
    ClusterStatus,
    CredentialsConfig,
    DisksConfig,
    DNSConfig,
    DockerRegistryConfig,
    EFSPerformanceMode,
    EFSThroughputMode,
    EMCECSCredentials,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GoogleCloudProvider,
    GoogleFilestoreTier,
    GoogleStorage,
    GrafanaCredentials,
    HelmRegistryConfig,
    IdleJobConfig,
    IngressConfig,
    IntelGPU,
    IntelGPUPreset,
    KubernetesCredentials,
    MetricsConfig,
    MinioCredentials,
    MonitoringConfig,
    NeuroAuthConfig,
    NodePool,
    NodePoolOptions,
    NvidiaGPU,
    NvidiaGPUPreset,
    OnPremCloudProvider,
    OpenStackCredentials,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchNodePoolResourcesRequest,
    PatchNodePoolSizeRequest,
    PatchOrchestratorConfigRequest,
    PrometheusCredentials,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    SentryCredentials,
    StorageConfig,
    StorageInstance,
    TPUPreset,
    TPUResource,
    VCDCloudProvider,
    VCDCloudProviderOptions,
    VCDCredentials,
    VCDStorage,
    VolumeConfig,
)
from neuro_config_client.factories import EntityFactory, PayloadFactory

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    # why not backports.zoneinfo: https://github.com/pganssle/zoneinfo/issues/125
    from backports.zoneinfo._zoneinfo import ZoneInfo


@pytest.fixture()
def nvidia_small_gpu() -> str:
    return "nvidia-tesla-k80"


@pytest.fixture()
def amd_small_gpu() -> str:
    return "instinct-mi25"


@pytest.fixture()
def intel_small_gpu() -> str:
    return "flex-170"


class TestEntityFactory:
    @pytest.fixture
    def factory(self) -> EntityFactory:
        return EntityFactory()

    def test_create_cluster_defaults(self, factory: EntityFactory) -> None:
        result = factory.create_cluster(
            {
                "name": "default",
                "status": "blank",
                "orchestrator": {
                    "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                    "job_fallback_hostname": "default.jobs-dev.neu.ro",
                    "job_schedule_timeout_s": 1,
                    "job_schedule_scale_up_timeout_s": 2,
                    "is_http_ingress_secure": False,
                    "resource_pool_types": [{"name": "node-pool"}],
                },
                "storage": {"url": "https://storage-dev.neu.ro"},
                "registry": {
                    "url": "https://registry-dev.neu.ro",
                    "email": "dev@neu.ro",
                },
                "monitoring": {"url": "https://monitoring-dev.neu.ro"},
                "secrets": {"url": "https://secrets-dev.neu.ro"},
                "metrics": {"url": "https://secrets-dev.neu.ro"},
                "disks": {
                    "url": "https://secrets-dev.neu.ro",
                    "storage_limit_per_user": 1024,
                },
                "ingress": {"acme_environment": "production"},
                "dns": {
                    "name": "neu.ro",
                    "a_records": [
                        {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}
                    ],
                },
                "buckets": {
                    "url": "https://buckets-dev.neu.ro",
                    "disable_creation": True,
                },
                "energy": {
                    "co2_grams_eq_per_kwh": 100,
                    "schedules": [
                        {"name": "default", "price_per_kwh": "123.4"},
                        {
                            "name": "green",
                            "price_per_kwh": "123.4",
                            "periods": [
                                {
                                    "weekday": 1,
                                    "start_time": "23:00",
                                    "end_time": "23:59",
                                }
                            ],
                        },
                    ],
                },
                "apps": {
                    "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
                    "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
                },
                "created_at": str(datetime.now()),
            }
        )

        assert result.timezone == ZoneInfo("UTC")
        assert result.location is None
        assert result.logo_url is None

    def test_create_cluster(
        self,
        factory: EntityFactory,
        google_cloud_provider_response: dict[str, Any],
        credentials: dict[str, Any],
    ) -> None:
        result = factory.create_cluster(
            {
                "name": "default",
                "status": "blank",
                "location": "us",
                "logo_url": "https://logo",
                "timezone": "America/Los_Angeles",
                "orchestrator": {
                    "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                    "job_fallback_hostname": "default.jobs-dev.neu.ro",
                    "job_schedule_timeout_s": 1,
                    "job_schedule_scale_up_timeout_s": 2,
                    "is_http_ingress_secure": False,
                    "resource_pool_types": [{"name": "node-pool"}],
                },
                "storage": {"url": "https://storage-dev.neu.ro"},
                "registry": {
                    "url": "https://registry-dev.neu.ro",
                    "email": "dev@neu.ro",
                },
                "monitoring": {"url": "https://monitoring-dev.neu.ro"},
                "secrets": {"url": "https://secrets-dev.neu.ro"},
                "metrics": {"url": "https://secrets-dev.neu.ro"},
                "disks": {
                    "url": "https://secrets-dev.neu.ro",
                    "storage_limit_per_user": 1024,
                },
                "ingress": {"acme_environment": "production"},
                "dns": {
                    "name": "neu.ro",
                    "a_records": [
                        {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}
                    ],
                },
                "buckets": {
                    "url": "https://buckets-dev.neu.ro",
                    "disable_creation": True,
                },
                "energy": {
                    "co2_grams_eq_per_kwh": 100,
                    "schedules": [
                        {"name": "default", "price_per_kwh": "123.4"},
                        {
                            "name": "green",
                            "price_per_kwh": "123.4",
                            "periods": [
                                {
                                    "weekday": 1,
                                    "start_time": "23:00",
                                    "end_time": "23:59",
                                }
                            ],
                        },
                    ],
                },
                "apps": {
                    "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
                    "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
                },
                "cloud_provider": google_cloud_provider_response,
                "credentials": credentials,
                "created_at": str(datetime.now()),
            }
        )

        assert result.name == "default"
        assert result.status == ClusterStatus.BLANK
        assert result.timezone == ZoneInfo("America/Los_Angeles")
        assert result.location == "us"
        assert result.logo_url == URL("https://logo")
        assert result.orchestrator
        assert result.storage
        assert result.registry
        assert result.monitoring
        assert result.secrets
        assert result.metrics
        assert result.disks
        assert result.ingress
        assert result.dns
        assert result.buckets
        assert result.energy
        assert result.apps
        assert result.cloud_provider
        assert result.credentials
        assert result.created_at

    def test_create_cluster__invalid_timezone(self, factory: EntityFactory) -> None:
        with pytest.raises(ValueError, match="invalid timezone"):
            factory.create_cluster(
                {
                    "name": "default",
                    "status": "blank",
                    "created_at": str(datetime.now()),
                    "timezone": "invalid",
                }
            )

    def test_create_orchestrator(self, factory: EntityFactory) -> None:
        result = factory.create_orchestrator(
            {
                "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                "job_fallback_hostname": "default.jobs-dev.neu.ro",
                "job_schedule_timeout_s": 1,
                "job_schedule_scale_up_timeout_s": 2,
                "is_http_ingress_secure": False,
                "resource_pool_types": [{"name": "node-pool"}],
                "resource_presets": [
                    {
                        "name": "cpu-micro",
                        "credits_per_hour": "10",
                        "cpu": 0.1,
                        "memory": 100,
                    }
                ],
                "allow_privileged_mode": True,
                "allow_job_priority": True,
                "pre_pull_images": ["neuromation/base"],
                "idle_jobs": [
                    {
                        "name": "idle",
                        "count": 1,
                        "image": "miner",
                        "resources": {
                            "cpu": 1,
                            "memory": 1024,
                            "nvidia_gpu": 1,
                            "amd_gpu": 2,
                            "intel_gpu": 3,
                        },
                    },
                    {
                        "name": "idle",
                        "count": 1,
                        "image": "miner",
                        "command": ["bash"],
                        "args": ["-c", "sleep infinity"],
                        "resources": {"cpu": 1, "memory": 1024},
                        "env": {"NAME": "VALUE"},
                        "node_selector": {"label": "value"},
                        "image_pull_secret": "secret",
                    },
                ],
            }
        )

        assert result == OrchestratorConfig(
            job_hostname_template="{job_id}.jobs-dev.neu.ro",
            job_fallback_hostname="default.jobs-dev.neu.ro",
            job_schedule_timeout_s=1,
            job_schedule_scale_up_timeout_s=2,
            is_http_ingress_secure=False,
            resource_pool_types=[mock.ANY],
            resource_presets=[mock.ANY],
            allow_privileged_mode=True,
            allow_job_priority=True,
            pre_pull_images=["neuromation/base"],
            idle_jobs=[
                IdleJobConfig(
                    name="idle",
                    count=1,
                    image="miner",
                    resources=Resources(
                        cpu=1,
                        memory=1024,
                        nvidia_gpu=1,
                        amd_gpu=2,
                        intel_gpu=3,
                    ),
                ),
                IdleJobConfig(
                    name="idle",
                    count=1,
                    image="miner",
                    command=["bash"],
                    args=["-c", "sleep infinity"],
                    resources=Resources(cpu=1, memory=1024),
                    env={"NAME": "VALUE"},
                    node_selector={"label": "value"},
                    image_pull_secret="secret",
                ),
            ],
        )

    def test_create_orchestrator_default(self, factory: EntityFactory) -> None:
        result = factory.create_orchestrator(
            {
                "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                "job_fallback_hostname": "default.jobs-dev.neu.ro",
                "job_schedule_timeout_s": 1,
                "job_schedule_scale_up_timeout_s": 2,
                "is_http_ingress_secure": False,
            }
        )

        assert result == OrchestratorConfig(
            job_hostname_template="{job_id}.jobs-dev.neu.ro",
            job_fallback_hostname="default.jobs-dev.neu.ro",
            job_schedule_timeout_s=1,
            job_schedule_scale_up_timeout_s=2,
            is_http_ingress_secure=False,
            resource_pool_types=[],
            resource_presets=[],
            idle_jobs=[],
        )

    def test_create_resource_pool_type(
        self,
        factory: EntityFactory,
        nvidia_small_gpu: str,
        amd_small_gpu: str,
        intel_small_gpu: str,
    ) -> None:
        result = factory.create_resource_pool_type(
            {
                "name": "n1-highmem-4",
                "min_size": 1,
                "max_size": 2,
                "idle_size": 1,
                "cpu": 4.0,
                "available_cpu": 3.0,
                "memory": 12 * 1024,
                "available_memory": 10 * 1024,
                "disk_size": 700,
                "nvidia_gpu": {
                    "count": 1,
                    "model": nvidia_small_gpu,
                    "memory": 123,
                },
                "amd_gpu": {
                    "count": 2,
                    "model": amd_small_gpu,
                },
                "intel_gpu": {
                    "count": 3,
                    "model": intel_small_gpu,
                },
                "tpu": {
                    "ipv4_cidr_block": "10.0.0.0/8",
                    "types": ["tpu"],
                    "software_versions": ["v1"],
                },
                "is_preemptible": True,
                "price": "1.0",
                "currency": "USD",
                "cpu_min_watts": 1.0,
                "cpu_max_watts": 2.0,
            }
        )

        assert result == ResourcePoolType(
            name="n1-highmem-4",
            min_size=1,
            max_size=2,
            idle_size=1,
            cpu=4.0,
            available_cpu=3.0,
            memory=12 * 1024,
            available_memory=10 * 1024,
            nvidia_gpu=NvidiaGPU(count=1, model=nvidia_small_gpu, memory=123),
            amd_gpu=AMDGPU(count=2, model=amd_small_gpu),
            intel_gpu=IntelGPU(count=3, model=intel_small_gpu),
            tpu=mock.ANY,
            is_preemptible=True,
            price=Decimal("1.0"),
            currency="USD",
            cpu_min_watts=1.0,
            cpu_max_watts=2.0,
        )

    def test_create_empty_resource_pool_type(self, factory: EntityFactory) -> None:
        result = factory.create_resource_pool_type({"name": "node-pool"})

        assert result == ResourcePoolType(name="node-pool")

    def test_create_tpu_resource(self, factory: EntityFactory) -> None:
        result = factory.create_tpu_resource(
            {
                "ipv4_cidr_block": "10.0.0.0/8",
                "types": ["tpu"],
                "software_versions": ["v1"],
            }
        )

        assert result == TPUResource(
            ipv4_cidr_block="10.0.0.0/8", types=["tpu"], software_versions=["v1"]
        )

    def test_create_resource_preset_default(self, factory: EntityFactory) -> None:
        result = factory.create_resource_preset(
            {
                "name": "cpu-small",
                "credits_per_hour": "10",
                "cpu": 4.0,
                "memory": 1024,
            }
        )

        assert result == ResourcePreset(
            name="cpu-small", credits_per_hour=Decimal("10"), cpu=4.0, memory=1024
        )

    def test_create_resource_preset_custom(
        self,
        factory: EntityFactory,
        nvidia_small_gpu: str,
    ) -> None:
        result = factory.create_resource_preset(
            {
                "name": "gpu-small",
                "credits_per_hour": "10",
                "cpu": 4.0,
                "memory": 12288,
                "nvidia_gpu": {
                    "count": 1,
                    "model": nvidia_small_gpu,
                    "memory": 123,
                },
                "amd_gpu": {"count": 2},
                "intel_gpu": {"count": 3},
                "tpu": {"type": "tpu", "software_version": "v1"},
                "scheduler_enabled": True,
                "preemptible_node": True,
                "resource_pool_names": ["gpu"],
                "available_resource_pool_names": ["available-gpu"],
                "is_external_job": True,
                "capacity": 10,
            }
        )

        assert result == ResourcePreset(
            name="gpu-small",
            credits_per_hour=Decimal("10"),
            cpu=4.0,
            memory=12288,
            nvidia_gpu=NvidiaGPUPreset(
                count=1,
                model=nvidia_small_gpu,
                memory=123,
            ),
            amd_gpu=AMDGPUPreset(count=2),
            intel_gpu=IntelGPUPreset(count=3),
            tpu=TPUPreset(type="tpu", software_version="v1"),
            scheduler_enabled=True,
            preemptible_node=True,
            resource_pool_names=["gpu"],
            available_resource_pool_names=["available-gpu"],
            is_external_job=True,
            capacity=10,
        )

    def test_create_storage(self, factory: EntityFactory) -> None:
        result = factory.create_storage({"url": "https://storage-dev.neu.ro"})

        assert result == StorageConfig(
            url=URL("https://storage-dev.neu.ro"), volumes=[]
        )

    def test_create_storage_with_volumes(self, factory: EntityFactory) -> None:
        result = factory.create_storage(
            {
                "url": "https://storage-dev.neu.ro",
                "volumes": [
                    {"name": "test-1"},
                    {
                        "name": "test-2",
                        "path": "/volume",
                        "size": 1024,
                        "credits_per_hour_per_gb": "123",
                    },
                ],
            }
        )

        assert result == StorageConfig(
            url=URL("https://storage-dev.neu.ro"),
            volumes=[
                VolumeConfig(name="test-1"),
                VolumeConfig(
                    name="test-2",
                    path="/volume",
                    size=1024,
                    credits_per_hour_per_gb=Decimal(123),
                ),
            ],
        )

    def test_create_registry(self, factory: EntityFactory) -> None:
        result = factory.create_registry({"url": "https://registry-dev.neu.ro"})

        assert result == RegistryConfig(url=URL("https://registry-dev.neu.ro"))

    def test_create_monitoring(self, factory: EntityFactory) -> None:
        result = factory.create_monitoring({"url": "https://monitoring-dev.neu.ro"})

        assert result == MonitoringConfig(url=URL("https://monitoring-dev.neu.ro"))

    def test_create_secrets(self, factory: EntityFactory) -> None:
        result = factory.create_secrets({"url": "https://secrets-dev.neu.ro"})

        assert result == SecretsConfig(url=URL("https://secrets-dev.neu.ro"))

    def test_create_metrics(self, factory: EntityFactory) -> None:
        result = factory.create_metrics({"url": "https://metrics-dev.neu.ro"})

        assert result == MetricsConfig(url=URL("https://metrics-dev.neu.ro"))

    def test_create_dns(self, factory: EntityFactory) -> None:
        result = factory.create_dns(
            {
                "name": "neu.ro",
                "a_records": [{"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}],
            }
        )

        assert result == DNSConfig(name="neu.ro", a_records=[mock.ANY])

    def test_create_a_record_with_ips(self, factory: EntityFactory) -> None:
        result = factory.create_a_record(
            {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}
        )

        assert result == ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])

    def test_create_a_record_dns_name(self, factory: EntityFactory) -> None:
        result = factory.create_a_record(
            {
                "name": "*.jobs-dev.neu.ro.",
                "dns_name": "load-balancer",
                "zone_id": "/hostedzone/1",
                "evaluate_target_health": True,
            }
        )

        assert result == ARecord(
            name="*.jobs-dev.neu.ro.",
            dns_name="load-balancer",
            zone_id="/hostedzone/1",
            evaluate_target_health=True,
        )

    def test_create_disks(self, factory: EntityFactory) -> None:
        result = factory.create_disks(
            {"url": "https://metrics-dev.neu.ro", "storage_limit_per_user": 1024}
        )

        assert result == DisksConfig(
            url=URL("https://metrics-dev.neu.ro"), storage_limit_per_user=1024
        )

    def test_create_buckets(self, factory: EntityFactory) -> None:
        result = factory.create_buckets(
            {"url": "https://buckets-dev.neu.ro", "disable_creation": True}
        )

        assert result == BucketsConfig(
            url=URL("https://buckets-dev.neu.ro"), disable_creation=True
        )

    def test_create_ingress(self, factory: EntityFactory) -> None:
        result = factory.create_ingress(
            {
                "acme_environment": "production",
                "default_cors_origins": ["https://console.apolo.us"],
                "additional_cors_origins": ["https://custom.app"],
            }
        )

        assert result == IngressConfig(
            acme_environment=ACMEEnvironment.PRODUCTION,
            default_cors_origins=["https://console.apolo.us"],
            additional_cors_origins=["https://custom.app"],
        )

    def test_create_ingress_defaults(self, factory: EntityFactory) -> None:
        result = factory.create_ingress({"acme_environment": "production"})

        assert result == IngressConfig(acme_environment=ACMEEnvironment.PRODUCTION)

    @pytest.fixture
    def google_cloud_provider_response(self) -> dict[str, Any]:
        return {
            "type": "gcp",
            "location_type": "zonal",
            "region": "us-central1",
            "zones": ["us-central1-a"],
            "project": "project",
            "credentials": {
                "type": "service_account",
                "project_id": "project_id",
                "private_key_id": "private_key_id",
                "private_key": "private_key",
                "client_email": "service.account@gmail.com",
                "client_id": "client_id",
                "auth_uri": "https://auth_uri",
                "token_uri": "https://token_uri",
                "auth_provider_x509_cert_url": "https://auth_provider_x509_cert_url",
                "client_x509_cert_url": "https://client_x509_cert_url",
            },
            "node_pools": [
                {
                    "name": "n1-highmem-8",
                    "role": "platform_job",
                    "machine_type": "n1-highmem-8",
                    "min_size": 0,
                    "max_size": 1,
                    "cpu": 8.0,
                    "available_cpu": 7.0,
                    "memory": 52 * 1024,
                    "available_memory": 45 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                },
                {
                    "name": "n1-highmem-32-1xk80-preemptible",
                    "role": "platform_job",
                    "machine_type": "n1-highmem-32",
                    "min_size": 0,
                    "max_size": 1,
                    "idle_size": 1,
                    "cpu": 32.0,
                    "available_cpu": 31.0,
                    "memory": 208 * 1024,
                    "available_memory": 201 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                    "nvidia_gpu": 1,
                    "nvidia_gpu_model": "nvidia-tesla-k80",
                    "is_preemptible": True,
                },
            ],
            "storage": {
                "description": "GCP Filestore (Premium)",
                "backend": "filestore",
                "tier": "PREMIUM",
                "instances": [
                    {"name": "test-1", "size": 5 * 1024 * 1024, "ready": False},
                    {"name": "test-2", "size": 3 * 1024 * 1024, "ready": True},
                ],
            },
        }

    @pytest.fixture
    def google_cloud_provider(self) -> GoogleCloudProvider:
        return GoogleCloudProvider(
            location_type=ClusterLocationType.ZONAL,
            region="us-central1",
            zones=["us-central1-a"],
            project="project",
            credentials={
                "type": "service_account",
                "project_id": "project_id",
                "private_key_id": "private_key_id",
                "private_key": "private_key",
                "client_email": "service.account@gmail.com",
                "client_id": "client_id",
                "auth_uri": "https://auth_uri",
                "token_uri": "https://token_uri",
                "auth_provider_x509_cert_url": "https://auth_provider_x509_cert_url",
                "client_x509_cert_url": "https://client_x509_cert_url",
            },
            node_pools=[
                NodePool(
                    name="n1-highmem-8",
                    machine_type="n1-highmem-8",
                    min_size=0,
                    max_size=1,
                    cpu=8.0,
                    available_cpu=7.0,
                    memory=52 * 1024,
                    available_memory=45 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                ),
                NodePool(
                    name="n1-highmem-32-1xk80-preemptible",
                    machine_type="n1-highmem-32",
                    min_size=0,
                    max_size=1,
                    idle_size=1,
                    cpu=32.0,
                    available_cpu=31.0,
                    memory=208 * 1024,
                    available_memory=201 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                    gpu=1,
                    gpu_model="nvidia-tesla-k80",
                    nvidia_gpu=1,
                    nvidia_gpu_model="nvidia-tesla-k80",
                    is_preemptible=True,
                ),
            ],
            storage=GoogleStorage(
                description="GCP Filestore (Premium)",
                tier=GoogleFilestoreTier.PREMIUM,
                instances=[
                    StorageInstance(name="test-1", size=5 * 1024 * 1024),
                    StorageInstance(name="test-2", size=3 * 1024 * 1024, ready=True),
                ],
            ),
        )

    def test_create_cloud_provider_google(
        self,
        factory: EntityFactory,
        google_cloud_provider: GoogleCloudProvider,
        google_cloud_provider_response: dict[str, Any],
    ) -> None:
        result = factory.create_cloud_provider(google_cloud_provider_response)
        assert result == google_cloud_provider

    @pytest.fixture
    def aws_cloud_provider_response(self) -> dict[str, Any]:
        return {
            "type": "aws",
            "region": "us-central-1",
            "zones": ["us-central-1a"],
            "vpc_id": "test-vpc",
            "credentials": {
                "access_key_id": "access_key_id",
                "secret_access_key": "secret_access_key",
            },
            "node_pools": [
                {
                    "role": "platform_job",
                    "name": "m5-2xlarge",
                    "machine_type": "m5.2xlarge",
                    "min_size": 0,
                    "max_size": 1,
                    "cpu": 8.0,
                    "available_cpu": 7.0,
                    "memory": 32 * 1024,
                    "available_memory": 28 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                },
                {
                    "role": "platform_job",
                    "name": "p2-xlarge-1xk80-preemptible",
                    "machine_type": "p2.xlarge",
                    "min_size": 0,
                    "max_size": 1,
                    "idle_size": 1,
                    "cpu": 4.0,
                    "available_cpu": 3.0,
                    "memory": 61 * 1024,
                    "available_memory": 57 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                    "nvidia_gpu": 1,
                    "nvidia_gpu_model": "nvidia-tesla-k80",
                    "is_preemptible": True,
                },
            ],
            "storage": {
                "description": "AWS EFS (generalPurpose, bursting)",
                "performance_mode": "generalPurpose",
                "throughput_mode": "bursting",
                "instances": [
                    {"name": "test-1", "ready": False},
                    {"name": "test-2", "ready": True},
                ],
            },
        }

    @pytest.fixture
    def aws_cloud_provider(self) -> AWSCloudProvider:
        return AWSCloudProvider(
            region="us-central-1",
            zones=["us-central-1a"],
            vpc_id="test-vpc",
            credentials=AWSCredentials(
                access_key_id="access_key_id", secret_access_key="secret_access_key"
            ),
            node_pools=[
                NodePool(
                    name="m5-2xlarge",
                    machine_type="m5.2xlarge",
                    min_size=0,
                    max_size=1,
                    cpu=8.0,
                    available_cpu=7.0,
                    memory=32 * 1024,
                    available_memory=28 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                ),
                NodePool(
                    name="p2-xlarge-1xk80-preemptible",
                    machine_type="p2.xlarge",
                    min_size=0,
                    max_size=1,
                    idle_size=1,
                    cpu=4.0,
                    available_cpu=3.0,
                    memory=61 * 1024,
                    available_memory=57 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                    gpu=1,
                    gpu_model="nvidia-tesla-k80",
                    nvidia_gpu=1,
                    nvidia_gpu_model="nvidia-tesla-k80",
                    is_preemptible=True,
                ),
            ],
            storage=AWSStorage(
                description="AWS EFS (generalPurpose, bursting)",
                performance_mode=EFSPerformanceMode.GENERAL_PURPOSE,
                throughput_mode=EFSThroughputMode.BURSTING,
                instances=[
                    StorageInstance(name="test-1"),
                    StorageInstance(name="test-2", ready=True),
                ],
            ),
        )

    def test_create_cloud_provider_aws(
        self,
        factory: EntityFactory,
        aws_cloud_provider: AWSCloudProvider,
        aws_cloud_provider_response: dict[str, Any],
    ) -> None:
        result = factory.create_cloud_provider(aws_cloud_provider_response)
        assert result == aws_cloud_provider

    @pytest.fixture
    def azure_cloud_provider_response(self) -> dict[str, Any]:
        return {
            "type": "azure",
            "region": "westus",
            "resource_group": "resource_group",
            "credentials": {
                "subscription_id": "subscription_id",
                "tenant_id": "tenant_id",
                "client_id": "client_id",
                "client_secret": "client_secret",
            },
            "node_pools": [
                {
                    "role": "platform_job",
                    "name": "Standard_D8s_v3",
                    "machine_type": "Standard_D8s_v3",
                    "min_size": 0,
                    "max_size": 1,
                    "cpu": 8.0,
                    "available_cpu": 7.0,
                    "memory": 32 * 1024,
                    "available_memory": 28 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                },
                {
                    "role": "platform_job",
                    "name": "Standard_NC6-1xk80-preemptible",
                    "machine_type": "Standard_NC6",
                    "min_size": 0,
                    "max_size": 1,
                    "idle_size": 1,
                    "cpu": 6.0,
                    "available_cpu": 5.0,
                    "memory": 56 * 1024,
                    "available_memory": 50 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                    "nvidia_gpu": 1,
                    "nvidia_gpu_model": "nvidia-tesla-k80",
                    "is_preemptible": True,
                },
            ],
            "storage": {
                "description": "Azure Files (Premium, LRS replication)",
                "tier": "Premium",
                "replication_type": "LRS",
                "instances": [
                    {"name": "test-1", "size": 100 * 1024, "ready": False},
                    {"name": "test-2", "size": 200 * 1024, "ready": True},
                ],
            },
        }

    @pytest.fixture
    def azure_cloud_provider(self) -> AzureCloudProvider:
        return AzureCloudProvider(
            region="westus",
            resource_group="resource_group",
            credentials=AzureCredentials(
                subscription_id="subscription_id",
                tenant_id="tenant_id",
                client_id="client_id",
                client_secret="client_secret",
            ),
            node_pools=[
                NodePool(
                    name="Standard_D8s_v3",
                    machine_type="Standard_D8s_v3",
                    min_size=0,
                    max_size=1,
                    cpu=8.0,
                    available_cpu=7.0,
                    memory=32 * 1024,
                    available_memory=28 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                ),
                NodePool(
                    name="Standard_NC6-1xk80-preemptible",
                    machine_type="Standard_NC6",
                    min_size=0,
                    max_size=1,
                    idle_size=1,
                    cpu=6.0,
                    available_cpu=5.0,
                    memory=56 * 1024,
                    available_memory=50 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                    gpu=1,
                    gpu_model="nvidia-tesla-k80",
                    nvidia_gpu=1,
                    nvidia_gpu_model="nvidia-tesla-k80",
                    is_preemptible=True,
                ),
            ],
            storage=AzureStorage(
                description="Azure Files (Premium, LRS replication)",
                tier=AzureStorageTier.PREMIUM,
                replication_type=AzureReplicationType.LRS,
                instances=[
                    StorageInstance(name="test-1", size=100 * 1024),
                    StorageInstance(name="test-2", size=200 * 1024, ready=True),
                ],
            ),
        )

    def test_create_cloud_provider_azure(
        self,
        factory: EntityFactory,
        azure_cloud_provider: AzureCloudProvider,
        azure_cloud_provider_response: dict[str, Any],
    ) -> None:
        result = factory.create_cloud_provider(azure_cloud_provider_response)
        assert result == azure_cloud_provider

    @pytest.fixture
    def on_prem_cloud_provider_response(self) -> dict[str, Any]:
        return {
            "type": "on_prem",
            "kubernetes_url": "localhost:8001",
            "credentials": {
                "token": "kubernetes-token",
                "ca_data": "kubernetes-ca-data",
            },
            "node_pools": [
                {
                    "role": "platform_job",
                    "min_size": 1,
                    "max_size": 1,
                    "name": "cpu-machine",
                    "machine_type": "cpu-machine",
                    "cpu": 1.0,
                    "available_cpu": 1.0,
                    "memory": 1024,
                    "available_memory": 1024,
                    "disk_size": 700,
                },
                {
                    "role": "platform_job",
                    "min_size": 1,
                    "max_size": 1,
                    "name": "gpu-machine-1xk80",
                    "machine_type": "gpu-machine-1xk80",
                    "cpu": 1.0,
                    "available_cpu": 1.0,
                    "memory": 1024,
                    "available_memory": 1024,
                    "disk_size": 700,
                    "nvidia_gpu": 1,
                    "nvidia_gpu_model": "nvidia-tesla-k80",
                    "price": "0.9",
                    "currency": "USD",
                    "cpu_min_watts": 0.1,
                    "cpu_max_watts": 100,
                },
            ],
        }

    @pytest.fixture
    def on_prem_cloud_provider(self) -> OnPremCloudProvider:
        return OnPremCloudProvider(
            kubernetes_url=URL("localhost:8001"),
            credentials=KubernetesCredentials(
                token="kubernetes-token", ca_data="kubernetes-ca-data"
            ),
            node_pools=[
                NodePool(
                    min_size=1,
                    max_size=1,
                    name="cpu-machine",
                    cpu=1.0,
                    available_cpu=1.0,
                    memory=1024,
                    available_memory=1024,
                    disk_size=700,
                    available_disk_size=700,
                    machine_type="cpu-machine",
                ),
                NodePool(
                    min_size=1,
                    max_size=1,
                    name="gpu-machine-1xk80",
                    cpu=1.0,
                    available_cpu=1.0,
                    memory=1024,
                    available_memory=1024,
                    disk_size=700,
                    available_disk_size=700,
                    gpu=1,
                    gpu_model="nvidia-tesla-k80",
                    nvidia_gpu=1,
                    nvidia_gpu_model="nvidia-tesla-k80",
                    price=Decimal("0.9"),
                    currency="USD",
                    machine_type="gpu-machine-1xk80",
                    cpu_min_watts=0.1,
                    cpu_max_watts=100,
                ),
            ],
            storage=None,
        )

    def test_create_cloud_provider_on_prem(
        self,
        factory: EntityFactory,
        on_prem_cloud_provider: OnPremCloudProvider,
        on_prem_cloud_provider_response: dict[str, Any],
    ) -> None:
        result = factory.create_cloud_provider(on_prem_cloud_provider_response)
        assert result == on_prem_cloud_provider

    @pytest.fixture
    def vcd_cloud_provider_response(self) -> dict[str, Any]:
        return {
            "type": "vcd_mts",
            "url": "vcd_url",
            "organization": "vcd_org",
            "virtual_data_center": "vdc",
            "edge_name": "edge",
            "edge_external_network_name": "edge-external-network",
            "edge_public_ip": "10.0.0.1",
            "catalog_name": "catalog",
            "credentials": {
                "user": "vcd_user",
                "password": "vcd_password",
                "ssh_password": "ssh-password",
            },
            "node_pools": [
                {
                    "role": "platform_job",
                    "min_size": 1,
                    "max_size": 1,
                    "name": "Master-neuro",
                    "machine_type": "Master-neuro",
                    "cpu": 8.0,
                    "available_cpu": 7.0,
                    "memory": 32 * 1024,
                    "available_memory": 29 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                },
                {
                    "role": "platform_job",
                    "min_size": 1,
                    "max_size": 1,
                    "name": "X16-neuro-1xk80",
                    "machine_type": "X16-neuro",
                    "cpu": 16.0,
                    "available_cpu": 15.0,
                    "memory": 40 * 1024,
                    "available_memory": 37 * 1024,
                    "disk_size": 700,
                    "available_disk_size": 670,
                    "nvidia_gpu": 1,
                    "nvidia_gpu_model": "nvidia-tesla-k80",
                    "price": "0.9",
                    "currency": "USD",
                    "cpu_min_watts": 0.1,
                    "cpu_max_watts": 100,
                },
            ],
            "storage": {
                "profile_name": "profile",
                "size": 10,
                "instances": [
                    {"name": "test-1", "size": 7 * 1024, "ready": False},
                    {"name": "test-2", "size": 3 * 1024, "ready": True},
                ],
                "description": "profile",
            },
        }

    @pytest.fixture
    def vcd_cloud_provider(self) -> VCDCloudProvider:
        return VCDCloudProvider(
            _type=CloudProviderType.VCD_MTS,
            url=URL("vcd_url"),
            organization="vcd_org",
            virtual_data_center="vdc",
            edge_name="edge",
            edge_external_network_name="edge-external-network",
            edge_public_ip="10.0.0.1",
            catalog_name="catalog",
            credentials=VCDCredentials(
                user="vcd_user", password="vcd_password", ssh_password="ssh-password"
            ),
            node_pools=[
                NodePool(
                    min_size=1,
                    max_size=1,
                    name="Master-neuro",
                    machine_type="Master-neuro",
                    cpu=8.0,
                    available_cpu=7.0,
                    memory=32 * 1024,
                    available_memory=29 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                ),
                NodePool(
                    min_size=1,
                    max_size=1,
                    name="X16-neuro-1xk80",
                    machine_type="X16-neuro",
                    cpu=16.0,
                    available_cpu=15.0,
                    memory=40 * 1024,
                    available_memory=37 * 1024,
                    disk_size=700,
                    available_disk_size=670,
                    gpu=1,
                    gpu_model="nvidia-tesla-k80",
                    nvidia_gpu=1,
                    nvidia_gpu_model="nvidia-tesla-k80",
                    price=Decimal("0.9"),
                    currency="USD",
                    cpu_min_watts=0.1,
                    cpu_max_watts=100,
                ),
            ],
            storage=VCDStorage(
                description="profile",
                profile_name="profile",
                size=10,
                instances=[
                    StorageInstance(name="test-1", size=7 * 1024),
                    StorageInstance(name="test-2", size=3 * 1024, ready=True),
                ],
            ),
        )

    def test_create_cloud_provider_vcd(
        self,
        factory: EntityFactory,
        vcd_cloud_provider: VCDCloudProvider,
        vcd_cloud_provider_response: dict[str, Any],
    ) -> None:
        result = factory.create_cloud_provider(vcd_cloud_provider_response)
        assert result == vcd_cloud_provider

    @pytest.fixture
    def credentials(self) -> dict[str, Any]:
        return {
            "neuro": {
                "url": "https://neu.ro",
                "token": "cluster_token",
            },
            "neuro_registry": {
                "url": "https://ghcr.io/neuro-inc",
                "username": "username",
                "password": "password",
                "email": "username@neu.ro",
            },
            "neuro_helm": {
                "url": "oci://neuro-inc.ghcr.io",
                "username": "username",
                "password": "password",
            },
            "grafana": {
                "username": "grafana-username",
                "password": "grafana-password",
            },
            "prometheus": {
                "username": "prometheus-username",
                "password": "prometheus-password",
            },
            "sentry": {
                "client_key_id": "key",
                "public_dsn": "dsn",
                "sample_rate": 0.2,
            },
            "docker_hub": {
                "url": "https://index.docker.io/v1",
                "username": "test",
                "password": "password",
                "email": "test@neu.ro",
            },
            "minio": {
                "username": "test",
                "password": "password",
            },
            "emc_ecs": {
                "access_key_id": "key_id",
                "secret_access_key": "secret_key",
                "s3_endpoint": "https://emc-ecs.s3",
                "management_endpoint": "https://emc-ecs.management",
                "s3_assumable_role": "s3-role",
            },
            "open_stack": {
                "account_id": "id",
                "password": "password",
                "s3_endpoint": "https://os.s3",
                "endpoint": "https://os.management",
                "region_name": "region",
            },
        }

    def test_create_credentials(
        self, factory: EntityFactory, credentials: dict[str, Any]
    ) -> None:
        result = factory.create_credentials(credentials)

        assert result == CredentialsConfig(
            neuro=NeuroAuthConfig(
                url=URL("https://neu.ro"),
                token="cluster_token",
            ),
            neuro_registry=DockerRegistryConfig(
                url=URL("https://ghcr.io/neuro-inc"),
                username="username",
                password="password",
                email="username@neu.ro",
            ),
            neuro_helm=HelmRegistryConfig(
                url=URL("oci://neuro-inc.ghcr.io"),
                username="username",
                password="password",
            ),
            grafana=GrafanaCredentials(
                username="grafana-username",
                password="grafana-password",
            ),
            prometheus=PrometheusCredentials(
                username="prometheus-username",
                password="prometheus-password",
            ),
            sentry=SentryCredentials(
                client_key_id="key", public_dsn=URL("dsn"), sample_rate=0.2
            ),
            docker_hub=DockerRegistryConfig(
                url=URL("https://index.docker.io/v1"),
                username="test",
                password="password",
                email="test@neu.ro",
            ),
            minio=MinioCredentials(
                username="test",
                password="password",
            ),
            emc_ecs=EMCECSCredentials(
                access_key_id="key_id",
                secret_access_key="secret_key",
                s3_endpoint=URL("https://emc-ecs.s3"),
                management_endpoint=URL("https://emc-ecs.management"),
                s3_assumable_role="s3-role",
            ),
            open_stack=OpenStackCredentials(
                account_id="id",
                password="password",
                s3_endpoint=URL("https://os.s3"),
                endpoint=URL("https://os.management"),
                region_name="region",
            ),
        )

    def test_create_minimal_credentials(
        self, factory: EntityFactory, credentials: dict[str, Any]
    ) -> None:
        del credentials["grafana"]
        del credentials["prometheus"]
        del credentials["sentry"]
        del credentials["docker_hub"]
        del credentials["minio"]
        del credentials["emc_ecs"]
        del credentials["open_stack"]
        result = factory.create_credentials(credentials)

        assert result == CredentialsConfig(
            neuro=NeuroAuthConfig(
                url=URL("https://neu.ro"),
                token="cluster_token",
            ),
            neuro_registry=DockerRegistryConfig(
                url=URL("https://ghcr.io/neuro-inc"),
                username="username",
                password="password",
                email="username@neu.ro",
            ),
            neuro_helm=HelmRegistryConfig(
                url=URL("oci://neuro-inc.ghcr.io"),
                username="username",
                password="password",
            ),
        )

    @pytest.fixture
    def node_pool_options_response(self) -> dict[str, Any]:
        return {
            "machine_type": "Standard_ND24s",
            "cpu": 24,
            "available_cpu": 23,
            "memory": 458752,
            "available_memory": 452608,
            "gpu": 4,
            "gpu_model": "nvidia-tesla-p40",
            "extra_info": "will be ignored",
        }

    @pytest.fixture
    def node_pool_options(self) -> NodePoolOptions:
        return NodePoolOptions(
            machine_type="Standard_ND24s",
            cpu=24,
            available_cpu=23,
            memory=458752,
            available_memory=452608,
            nvidia_gpu=4,
            nvidia_gpu_model="nvidia-tesla-p40",
        )

    def test_aws_cloud_provider_options(
        self,
        factory: EntityFactory,
        node_pool_options_response: dict[str, Any],
        node_pool_options: NodePoolOptions,
    ) -> None:
        response = {
            "node_pools": [node_pool_options_response],
        }
        result = factory.create_cloud_provider_options(CloudProviderType.AWS, response)

        assert result == CloudProviderOptions(
            type=CloudProviderType.AWS,
            node_pools=[node_pool_options],
        )

    def test_aws_cloud_provider_options_defaults(self, factory: EntityFactory) -> None:
        result = factory.create_cloud_provider_options(CloudProviderType.AWS, {})

        assert result == CloudProviderOptions(type=CloudProviderType.AWS, node_pools=[])

    def test_google_cloud_provider_options(
        self,
        factory: EntityFactory,
        node_pool_options_response: dict[str, Any],
        node_pool_options: NodePoolOptions,
    ) -> None:
        response = {
            "node_pools": [node_pool_options_response],
        }
        result = factory.create_cloud_provider_options(CloudProviderType.GCP, response)

        assert result == CloudProviderOptions(
            type=CloudProviderType.GCP, node_pools=[node_pool_options]
        )

    def test_azure_cloud_provider_options(
        self,
        factory: EntityFactory,
        node_pool_options_response: dict[str, Any],
        node_pool_options: NodePoolOptions,
    ) -> None:
        response = {
            "node_pools": [node_pool_options_response],
        }
        result = factory.create_cloud_provider_options(
            CloudProviderType.AZURE, response
        )

        assert result == CloudProviderOptions(
            type=CloudProviderType.AZURE, node_pools=[node_pool_options]
        )

    def test_vcd_cloud_provider_options_defaults(
        self,
        factory: EntityFactory,
        node_pool_options_response: dict[str, Any],
        node_pool_options: NodePoolOptions,
    ) -> None:
        response = {
            "node_pools": [node_pool_options_response],
        }
        result = factory.create_cloud_provider_options(
            CloudProviderType.VCD_MTS, response
        )

        assert result == VCDCloudProviderOptions(
            type=CloudProviderType.VCD_MTS,
            node_pools=[node_pool_options],
        )

    def test_vcd_cloud_provider_options(
        self,
        factory: EntityFactory,
        node_pool_options_response: dict[str, Any],
        node_pool_options: NodePoolOptions,
    ) -> None:
        response = {
            "node_pools": [node_pool_options_response],
            "url": "https://vcd",
            "organization": "neuro-org",
            "edge_name_template": "neuro-edge",
            "edge_external_network_name": "neuro-edge-external",
            "catalog_name": "neuro",
            "storage_profile_names": ["neuro-storage"],
        }
        result = factory.create_cloud_provider_options(
            CloudProviderType.VCD_MTS, response
        )

        assert result == VCDCloudProviderOptions(
            type=CloudProviderType.VCD_MTS,
            node_pools=[node_pool_options],
            url=URL("https://vcd"),
            organization="neuro-org",
            edge_name_template="neuro-edge",
            edge_external_network_name="neuro-edge-external",
            catalog_name="neuro",
            storage_profile_names=["neuro-storage"],
        )

    def test_create_energy(self, factory: EntityFactory) -> None:
        timezone = ZoneInfo("America/Los_Angeles")
        energy = factory.create_energy(
            {
                "co2_grams_eq_per_kwh": 100,
                "schedules": [
                    {"name": "default", "price_per_kwh": "123.4"},
                    {
                        "name": "green",
                        "price_per_kwh": "123.4",
                        "periods": [
                            {"weekday": 1, "start_time": "23:00", "end_time": "23:59"}
                        ],
                    },
                ],
            },
            timezone=timezone,
        )

        assert energy == EnergyConfig(
            co2_grams_eq_per_kwh=100,
            schedules=[
                EnergySchedule(
                    name="default",
                    price_per_kwh=Decimal("123.4"),
                    periods=[],
                ),
                EnergySchedule(
                    name="green",
                    price_per_kwh=Decimal("123.4"),
                    periods=[
                        EnergySchedulePeriod(
                            weekday=1,
                            start_time=time(hour=23, minute=0, tzinfo=timezone),
                            end_time=time(hour=23, minute=59, tzinfo=timezone),
                        )
                    ],
                ),
            ],
        )

    @pytest.fixture
    def apps(self) -> AppsConfig:
        return AppsConfig(
            apps_hostname_templates=["{app_name}.apps.default.org.neu.ro"],
            app_proxy_url=URL("outputs-proxy.apps.default.org.neu.ro"),
        )

    @pytest.fixture
    def apps_dict(self) -> dict[str, Any]:
        return {
            "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
            "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
        }

    def test_create_apps(
        self, factory: EntityFactory, apps_dict: dict[str, Any], apps: AppsConfig
    ) -> None:
        result = factory.create_apps(apps_dict)
        assert result == apps


class TestPayloadFactory:
    @pytest.fixture
    def factory(self) -> PayloadFactory:
        return PayloadFactory()

    def test_create_patch_cluster_request(
        self, factory: PayloadFactory, credentials: CredentialsConfig
    ) -> None:
        result = factory.create_patch_cluster_request(
            PatchClusterRequest(
                credentials=credentials,
                location="us",
                logo_url=URL("https://logo"),
                storage=StorageConfig(url=URL("https://storage-dev.neu.ro")),
                registry=RegistryConfig(url=URL("https://registry-dev.neu.ro")),
                orchestrator=PatchOrchestratorConfigRequest(),
                monitoring=MonitoringConfig(url=URL("https://monitoring-dev.neu.ro")),
                secrets=SecretsConfig(url=URL("https://secrets-dev.neu.ro")),
                metrics=MetricsConfig(url=URL("https://metrics-dev.neu.ro")),
                disks=DisksConfig(
                    url=URL("https://metrics-dev.neu.ro"), storage_limit_per_user=1024
                ),
                buckets=BucketsConfig(
                    url=URL("https://buckets-dev.neu.ro"), disable_creation=True
                ),
                ingress=IngressConfig(acme_environment=ACMEEnvironment.PRODUCTION),
                dns=DNSConfig(
                    name="neu.ro",
                    a_records=[ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])],
                ),
                timezone=ZoneInfo("America/Los_Angeles"),
                energy=EnergyConfig(co2_grams_eq_per_kwh=100),
                apps=AppsConfig(
                    apps_hostname_templates=["{app_name}.apps.default.org.neu.ro"],
                    app_proxy_url=URL("outputs-proxy.apps.default.org.neu.ro"),
                ),
            )
        )

        assert result == {
            "location": "us",
            "logo_url": "https://logo",
            "credentials": mock.ANY,
            "storage": mock.ANY,
            "registry": mock.ANY,
            "orchestrator": mock.ANY,
            "monitoring": mock.ANY,
            "secrets": mock.ANY,
            "metrics": mock.ANY,
            "disks": mock.ANY,
            "buckets": mock.ANY,
            "ingress": mock.ANY,
            "dns": mock.ANY,
            "timezone": "America/Los_Angeles",
            "energy": mock.ANY,
            "apps": mock.ANY,
        }

    def test_create_patch_cluster_request_default(
        self, factory: PayloadFactory
    ) -> None:
        result = factory.create_patch_cluster_request(PatchClusterRequest())

        assert result == {}

    def test_create_orchestrator(self, factory: PayloadFactory) -> None:
        result = factory.create_orchestrator(
            OrchestratorConfig(
                job_hostname_template="{job_id}.jobs-dev.neu.ro",
                job_fallback_hostname="default.jobs-dev.neu.ro",
                job_schedule_timeout_s=1,
                job_schedule_scale_up_timeout_s=2,
                is_http_ingress_secure=False,
                allow_privileged_mode=True,
                allow_job_priority=True,
                resource_pool_types=[ResourcePoolType(name="cpu")],
                resource_presets=[
                    ResourcePreset(
                        name="cpu-micro",
                        credits_per_hour=Decimal(10),
                        cpu=0.1,
                        memory=100,
                    )
                ],
                pre_pull_images=["neuromation/base"],
                idle_jobs=[
                    IdleJobConfig(
                        name="idle",
                        count=1,
                        image="miner",
                        resources=Resources(
                            cpu=1,
                            memory=1024,
                            nvidia_gpu=1,
                            amd_gpu=2,
                            intel_gpu=3,
                        ),
                    ),
                    IdleJobConfig(
                        name="idle",
                        count=1,
                        image="miner",
                        command=["bash"],
                        args=["-c", "sleep infinity"],
                        resources=Resources(cpu=1, memory=1024),
                        env={"NAME": "VALUE"},
                        node_selector={"label": "value"},
                        image_pull_secret="secret",
                    ),
                ],
            )
        )

        assert result == {
            "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
            "job_fallback_hostname": "default.jobs-dev.neu.ro",
            "job_schedule_timeout_s": 1,
            "job_schedule_scale_up_timeout_s": 2,
            "is_http_ingress_secure": False,
            "resource_pool_types": [mock.ANY],
            "resource_presets": [mock.ANY],
            "allow_privileged_mode": True,
            "allow_job_priority": True,
            "pre_pull_images": ["neuromation/base"],
            "idle_jobs": [
                {
                    "name": "idle",
                    "count": 1,
                    "image": "miner",
                    "resources": {
                        "cpu": 1,
                        "memory": 1024,
                        "nvidia_gpu": 1,
                        "amd_gpu": 2,
                        "intel_gpu": 3,
                    },
                },
                {
                    "name": "idle",
                    "count": 1,
                    "image": "miner",
                    "command": ["bash"],
                    "args": ["-c", "sleep infinity"],
                    "resources": {"cpu": 1, "memory": 1024},
                    "env": {"NAME": "VALUE"},
                    "node_selector": {"label": "value"},
                    "image_pull_secret": "secret",
                },
            ],
        }

    def test_create_orchestrator_default(self, factory: PayloadFactory) -> None:
        result = factory.create_orchestrator(
            OrchestratorConfig(
                job_hostname_template="{job_id}.jobs-dev.neu.ro",
                job_fallback_hostname="default.jobs-dev.neu.ro",
                job_schedule_timeout_s=1,
                job_schedule_scale_up_timeout_s=2,
                is_http_ingress_secure=False,
            )
        )

        assert result == {
            "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
            "job_fallback_hostname": "default.jobs-dev.neu.ro",
            "job_schedule_timeout_s": 1,
            "job_schedule_scale_up_timeout_s": 2,
            "is_http_ingress_secure": False,
        }

    def test_create_patch_orchestrator_request(self, factory: PayloadFactory) -> None:
        result = factory.create_patch_orchestrator_request(
            PatchOrchestratorConfigRequest(
                job_hostname_template="{job_id}.jobs-dev.neu.ro",
                job_fallback_hostname="default.jobs-dev.neu.ro",
                job_schedule_timeout_s=1,
                job_schedule_scale_up_timeout_s=2,
                is_http_ingress_secure=False,
                allow_privileged_mode=True,
                allow_job_priority=True,
                resource_pool_types=[ResourcePoolType(name="cpu")],
                resource_presets=[
                    ResourcePreset(
                        name="cpu-micro",
                        credits_per_hour=Decimal(10),
                        cpu=0.1,
                        memory=100,
                    )
                ],
                pre_pull_images=["neuromation/base"],
                idle_jobs=[
                    IdleJobConfig(
                        name="idle",
                        count=1,
                        image="miner",
                        resources=Resources(cpu=1, memory=1024),
                    )
                ],
            )
        )

        assert result == {
            "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
            "job_fallback_hostname": "default.jobs-dev.neu.ro",
            "job_schedule_timeout_s": 1,
            "job_schedule_scale_up_timeout_s": 2,
            "is_http_ingress_secure": False,
            "resource_pool_types": [mock.ANY],
            "resource_presets": [mock.ANY],
            "allow_privileged_mode": True,
            "allow_job_priority": True,
            "pre_pull_images": ["neuromation/base"],
            "idle_jobs": [
                {
                    "name": "idle",
                    "count": 1,
                    "image": "miner",
                    "resources": {"cpu": 1, "memory": 1024},
                }
            ],
        }

    def test_create_patch_orchestrator_request_default(
        self, factory: PayloadFactory
    ) -> None:
        result = factory.create_patch_orchestrator_request(
            PatchOrchestratorConfigRequest()
        )

        assert result == {}

    def test_create_resource_pool_type(
        self,
        factory: PayloadFactory,
        nvidia_small_gpu: str,
        amd_small_gpu: str,
        intel_small_gpu: str,
    ) -> None:
        result = factory.create_resource_pool_type(
            ResourcePoolType(
                name="n1-highmem-4",
                min_size=1,
                max_size=2,
                idle_size=1,
                cpu=4.0,
                available_cpu=3.0,
                memory=12 * 1024,
                available_memory=10 * 1024,
                disk_size=700,
                available_disk_size=670,
                nvidia_gpu=NvidiaGPU(count=1, model=nvidia_small_gpu, memory=123),
                amd_gpu=AMDGPU(count=2, model=amd_small_gpu),
                intel_gpu=IntelGPU(count=3, model=intel_small_gpu),
                tpu=TPUResource(
                    ipv4_cidr_block="10.0.0.0/8",
                    types=["tpu"],
                    software_versions=["v1"],
                ),
                is_preemptible=True,
                price=Decimal("1.0"),
                currency="USD",
                cpu_min_watts=1.0,
                cpu_max_watts=2.0,
            )
        )

        assert result == {
            "name": "n1-highmem-4",
            "min_size": 1,
            "max_size": 2,
            "idle_size": 1,
            "cpu": 4.0,
            "available_cpu": 3.0,
            "memory": 12 * 1024,
            "available_memory": 10 * 1024,
            "disk_size": 700,
            "available_disk_size": 670,
            "nvidia_gpu": {
                "count": 1,
                "model": nvidia_small_gpu,
                "memory": 123,
            },
            "amd_gpu": 2,
            "amd_gpu_model": amd_small_gpu,
            "intel_gpu": 3,
            "intel_gpu_model": intel_small_gpu,
            "tpu": {
                "ipv4_cidr_block": "10.0.0.0/8",
                "types": ["tpu"],
                "software_versions": ["v1"],
            },
            "is_preemptible": True,
            "price": "1.0",
            "currency": "USD",
            "cpu_min_watts": 1.0,
            "cpu_max_watts": 2.0,
        }

    def test_create_empty_resource_pool_type(self, factory: PayloadFactory) -> None:
        result = factory.create_resource_pool_type(ResourcePoolType(name="node-pool"))

        assert result == {
            "name": "node-pool",
            "cpu": 1.0,
            "available_cpu": 1.0,
            "memory": 2**30,
            "available_memory": 2**30,
            "disk_size": 150 * 2**30,
            "available_disk_size": 150 * 2**30,
            "idle_size": 0,
            "is_preemptible": False,
            "min_size": 0,
            "max_size": 1,
        }

    def test_create_tpu_resource(self, factory: PayloadFactory) -> None:
        result = factory.create_tpu_resource(
            TPUResource(
                ipv4_cidr_block="10.0.0.0/8", types=["tpu"], software_versions=["v1"]
            )
        )

        assert result == {
            "ipv4_cidr_block": "10.0.0.0/8",
            "types": ["tpu"],
            "software_versions": ["v1"],
        }

    def test_create_resource_preset(self, factory: PayloadFactory) -> None:
        result = factory.create_resource_preset(
            ResourcePreset(
                name="cpu-small",
                credits_per_hour=Decimal("10"),
                cpu=4.0,
                memory=1024,
            )
        )

        assert result == {
            "name": "cpu-small",
            "credits_per_hour": "10",
            "cpu": 4.0,
            "memory": 1024,
        }

    def test_create_resource_preset__custom(
        self,
        factory: PayloadFactory,
        nvidia_small_gpu: str,
        amd_small_gpu: str,
    ) -> None:
        result = factory.create_resource_preset(
            ResourcePreset(
                name="gpu-small",
                credits_per_hour=Decimal("10"),
                cpu=4.0,
                memory=12288,
                nvidia_gpu=NvidiaGPUPreset(
                    count=1,
                    model=nvidia_small_gpu,
                    memory=123,
                ),
                amd_gpu=AMDGPUPreset(count=2, model=amd_small_gpu),
                intel_gpu=IntelGPUPreset(count=3),
                tpu=TPUPreset(type="tpu", software_version="v1"),
                scheduler_enabled=True,
                preemptible_node=True,
                resource_pool_names=["gpu"],
            )
        )

        assert result == {
            "name": "gpu-small",
            "credits_per_hour": "10",
            "cpu": 4.0,
            "memory": 12288,
            "nvidia_gpu": {
                "count": 1,
                "model": nvidia_small_gpu,
                "memory": 123,
            },
            "amd_gpu": 2,
            "amd_gpu_model": amd_small_gpu,
            "intel_gpu": 3,
            "tpu": {"type": "tpu", "software_version": "v1"},
            "scheduler_enabled": True,
            "preemptible_node": True,
            "resource_pool_names": ["gpu"],
        }

    def test_create_storage(self, factory: PayloadFactory) -> None:
        result = factory.create_storage(
            StorageConfig(url=URL("https://storage-dev.neu.ro"))
        )

        assert result == {"url": "https://storage-dev.neu.ro"}

    def test_create_storage_with_volumes(self, factory: PayloadFactory) -> None:
        result = factory.create_storage(
            StorageConfig(
                url=URL("https://storage-dev.neu.ro"),
                volumes=[
                    VolumeConfig(name="test-1"),
                    VolumeConfig(
                        name="test-2",
                        path="/volume",
                        size=1024,
                        credits_per_hour_per_gb=Decimal(123),
                    ),
                ],
            )
        )

        assert result == {
            "url": "https://storage-dev.neu.ro",
            "volumes": [
                {
                    "name": "test-1",
                    "credits_per_hour_per_gb": "0",
                },
                {
                    "name": "test-2",
                    "path": "/volume",
                    "size": 1024,
                    "credits_per_hour_per_gb": "123",
                },
            ],
        }

    def test_create_registry(self, factory: PayloadFactory) -> None:
        result = factory.create_registry(
            RegistryConfig(url=URL("https://registry-dev.neu.ro"))
        )

        assert result == {"url": "https://registry-dev.neu.ro"}

    def test_create_monitoring(self, factory: PayloadFactory) -> None:
        result = factory.create_monitoring(
            MonitoringConfig(url=URL("https://monitoring-dev.neu.ro"))
        )

        assert result == {"url": "https://monitoring-dev.neu.ro"}

    def test_create_secrets(self, factory: PayloadFactory) -> None:
        result = factory.create_secrets(
            SecretsConfig(url=URL("https://secrets-dev.neu.ro"))
        )

        assert result == {"url": "https://secrets-dev.neu.ro"}

    def test_create_metrics(self, factory: PayloadFactory) -> None:
        result = factory.create_metrics(
            MetricsConfig(url=URL("https://metrics-dev.neu.ro"))
        )

        assert result == {"url": "https://metrics-dev.neu.ro"}

    def test_create_dns(self, factory: PayloadFactory) -> None:
        result = factory.create_dns(
            DNSConfig(
                name="neu.ro",
                a_records=[ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])],
            )
        )

        assert result == {
            "name": "neu.ro",
            "a_records": [{"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}],
        }

    def test_create_a_record_with_ips(self, factory: PayloadFactory) -> None:
        result = factory.create_a_record(
            ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])
        )

        assert result == {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}

    def test_create_a_record_dns_name(self, factory: PayloadFactory) -> None:
        result = factory.create_a_record(
            ARecord(
                name="*.jobs-dev.neu.ro.",
                dns_name="load-balancer",
                zone_id="/hostedzone/1",
                evaluate_target_health=True,
            )
        )

        assert result == {
            "name": "*.jobs-dev.neu.ro.",
            "dns_name": "load-balancer",
            "zone_id": "/hostedzone/1",
            "evaluate_target_health": True,
        }

    def test_create_disks(self, factory: PayloadFactory) -> None:
        result = factory.create_disks(
            DisksConfig(
                url=URL("https://metrics-dev.neu.ro"), storage_limit_per_user=1024
            )
        )

        assert result == {
            "url": "https://metrics-dev.neu.ro",
            "storage_limit_per_user": 1024,
        }

    def test_create_buckets(self, factory: PayloadFactory) -> None:
        result = factory.create_buckets(
            BucketsConfig(url=URL("https://buckets-dev.neu.ro"), disable_creation=True)
        )

        assert result == {"url": "https://buckets-dev.neu.ro", "disable_creation": True}

    def test_create_ingress(self, factory: PayloadFactory) -> None:
        result = factory.create_ingress(
            IngressConfig(
                acme_environment=ACMEEnvironment.PRODUCTION,
                additional_cors_origins=["https://custom.app"],
            )
        )

        assert result == {
            "acme_environment": "production",
            "additional_cors_origins": ["https://custom.app"],
        }

    def test_create_ingress_defaults(self, factory: PayloadFactory) -> None:
        result = factory.create_ingress(
            IngressConfig(acme_environment=ACMEEnvironment.PRODUCTION)
        )

        assert result == {"acme_environment": "production"}

    @pytest.fixture
    def credentials(self) -> CredentialsConfig:
        return CredentialsConfig(
            neuro=NeuroAuthConfig(
                url=URL("https://neu.ro"),
                token="cluster_token",
            ),
            neuro_registry=DockerRegistryConfig(
                url=URL("https://ghcr.io/neuro-inc"),
                username="username",
                password="password",
                email="username@neu.ro",
            ),
            neuro_helm=HelmRegistryConfig(
                url=URL("oci://neuro-inc.ghcr.io"),
                username="username",
                password="password",
            ),
            grafana=GrafanaCredentials(
                username="grafana-username",
                password="grafana-password",
            ),
            prometheus=PrometheusCredentials(
                username="prometheus-username",
                password="prometheus-password",
            ),
            sentry=SentryCredentials(
                client_key_id="key", public_dsn=URL("dsn"), sample_rate=0.2
            ),
            docker_hub=DockerRegistryConfig(
                url=URL("https://index.docker.io/v1"),
                username="test",
                password="password",
                email="test@neu.ro",
            ),
            minio=MinioCredentials(
                username="test",
                password="password",
            ),
            emc_ecs=EMCECSCredentials(
                access_key_id="key_id",
                secret_access_key="secret_key",
                s3_endpoint=URL("https://emc-ecs.s3"),
                management_endpoint=URL("https://emc-ecs.management"),
                s3_assumable_role="s3-role",
            ),
            open_stack=OpenStackCredentials(
                account_id="id",
                password="password",
                s3_endpoint=URL("https://os.s3"),
                endpoint=URL("https://os.management"),
                region_name="region",
            ),
        )

    def test_create_credentials(
        self, factory: PayloadFactory, credentials: CredentialsConfig
    ) -> None:
        result = factory.create_credentials(credentials)

        assert result == {
            "neuro": {"token": "cluster_token"},
            "neuro_registry": {"username": "username", "password": "password"},
            "neuro_helm": {"username": "username", "password": "password"},
            "grafana": {
                "username": "grafana-username",
                "password": "grafana-password",
            },
            "prometheus": {
                "username": "prometheus-username",
                "password": "prometheus-password",
            },
            "sentry": {
                "client_key_id": "key",
                "public_dsn": "dsn",
                "sample_rate": 0.2,
            },
            "docker_hub": {"username": "test", "password": "password"},
            "minio": {
                "username": "test",
                "password": "password",
            },
            "emc_ecs": {
                "access_key_id": "key_id",
                "secret_access_key": "secret_key",
                "s3_endpoint": "https://emc-ecs.s3",
                "management_endpoint": "https://emc-ecs.management",
                "s3_assumable_role": "s3-role",
            },
            "open_stack": {
                "account_id": "id",
                "password": "password",
                "s3_endpoint": "https://os.s3",
                "endpoint": "https://os.management",
                "region_name": "region",
            },
        }

    def test_create_minimal_credentials(
        self, factory: PayloadFactory, credentials: CredentialsConfig
    ) -> None:
        credentials = replace(
            credentials,
            grafana=None,
            prometheus=None,
            sentry=None,
            docker_hub=None,
            minio=None,
            emc_ecs=None,
            open_stack=None,
        )
        result = factory.create_credentials(credentials)

        assert result == {
            "neuro": {"token": "cluster_token"},
            "neuro_registry": {"username": "username", "password": "password"},
            "neuro_helm": {"username": "username", "password": "password"},
        }

    @pytest.fixture
    def node_pool(self) -> NodePool:
        return NodePool(
            name="my-node-pool",
            min_size=0,
            max_size=10,
            idle_size=1,
            machine_type="some-machine-type",
            cpu=10,
            available_cpu=9,
            memory=2048,
            available_memory=1024,
            disk_size=100500,
            available_disk_size=100000,
            disk_type="some-disk-type",
            nvidia_gpu=1,
            nvidia_gpu_model="some-gpu-model",
            price=Decimal(180),
            currency="rabbits",
            is_preemptible=True,
            zones=("here", "there"),
            cpu_min_watts=0.01,
            cpu_max_watts=1000,
        )

    def test_create_add_node_pool_request(self, factory: PayloadFactory) -> None:
        node_pool = AddNodePoolRequest(
            name="my-node-pool",
            min_size=0,
            max_size=10,
            idle_size=1,
            machine_type="some-machine-type",
            cpu=10,
            available_cpu=9,
            memory=2048,
            available_memory=1024,
            disk_size=100500,
            available_disk_size=100000,
            disk_type="some-disk-type",
            nvidia_gpu=1,
            nvidia_gpu_model="some-gpu-model",
            price=Decimal(180),
            currency="rabbits",
            is_preemptible=True,
            zones=("here", "there"),
            cpu_min_watts=0.01,
            cpu_max_watts=1000,
        )
        payload = factory.create_add_node_pool_request(node_pool)

        assert payload == {
            "name": "my-node-pool",
            "role": "platform_job",
            "min_size": 0,
            "max_size": 10,
            "idle_size": 1,
            "is_preemptible": True,
            "machine_type": "some-machine-type",
            "cpu": 10,
            "available_cpu": 9,
            "memory": 2048,
            "available_memory": 1024,
            "disk_size": 100500,
            "available_disk_size": 100000,
            "disk_type": "some-disk-type",
            "nvidia_gpu": 1,
            "nvidia_gpu_model": "some-gpu-model",
            "zones": ("here", "there"),
            "price": "180",
            "currency": "rabbits",
            "cpu_min_watts": 0.01,
            "cpu_max_watts": 1000,
        }

    def test_create_add_node_pool_request_default(
        self, factory: PayloadFactory
    ) -> None:
        node_pool = AddNodePoolRequest(name="my-node-pool", min_size=0, max_size=1)

        payload = factory.create_add_node_pool_request(node_pool)

        assert payload == {
            "name": "my-node-pool",
            "role": "platform_job",
            "min_size": 0,
            "max_size": 1,
        }

    def test_create_patch_node_pool_size_request(self, factory: PayloadFactory) -> None:
        payload = factory.create_patch_node_pool_request(
            PatchNodePoolSizeRequest(min_size=1, max_size=3, idle_size=2)
        )

        assert payload == {
            "min_size": 1,
            "max_size": 3,
            "idle_size": 2,
        }

    def test_create_patch_node_pool_size_request_default(
        self, factory: PayloadFactory
    ) -> None:
        payload = factory.create_patch_node_pool_request(PatchNodePoolSizeRequest())

        assert payload == {}

    def test_create_patch_node_pool_resources_request(
        self, factory: PayloadFactory
    ) -> None:
        payload = factory.create_patch_node_pool_request(
            PatchNodePoolResourcesRequest(
                min_size=0,
                max_size=10,
                machine_type="n1-highmem-8",
                cpu=1,
                available_cpu=0.9,
                memory=1024,
                available_memory=1023,
                disk_size=100,
                available_disk_size=75,
                nvidia_gpu=1,
                nvidia_gpu_model="nvidia-gpu",
                amd_gpu=1,
                amd_gpu_model="amd-gpu",
                intel_gpu=1,
                intel_gpu_model="intel-gpu",
            )
        )

        assert payload == {
            "min_size": 0,
            "max_size": 10,
            "machine_type": "n1-highmem-8",
            "cpu": 1,
            "available_cpu": 0.9,
            "memory": 1024,
            "available_memory": 1023,
            "disk_size": 100,
            "available_disk_size": 75,
            "nvidia_gpu": 1,
            "nvidia_gpu_model": "nvidia-gpu",
            "amd_gpu": 1,
            "amd_gpu_model": "amd-gpu",
            "intel_gpu": 1,
            "intel_gpu_model": "intel-gpu",
        }

    def test_create_patch_node_pool_resources_request_default(
        self, factory: PayloadFactory
    ) -> None:
        payload = factory.create_patch_node_pool_request(
            PatchNodePoolResourcesRequest(
                cpu=1,
                available_cpu=0.9,
                memory=1024,
                available_memory=1023,
                disk_size=100,
                available_disk_size=75,
            )
        )

        assert payload == {
            "cpu": 1,
            "available_cpu": 0.9,
            "memory": 1024,
            "available_memory": 1023,
            "disk_size": 100,
            "available_disk_size": 75,
        }

    def test_create_energy(self, factory: PayloadFactory) -> None:
        timezone = ZoneInfo("America/Los_Angeles")
        energy = factory.create_energy(
            EnergyConfig(
                co2_grams_eq_per_kwh=100,
                schedules=[
                    EnergySchedule(name="default", price_per_kwh=Decimal("246.8")),
                    EnergySchedule(
                        name="green",
                        price_per_kwh=Decimal("123.4"),
                        periods=[
                            EnergySchedulePeriod(
                                weekday=1,
                                start_time=time(hour=23, minute=0, tzinfo=timezone),
                                end_time=time(hour=0, minute=0, tzinfo=timezone),
                            )
                        ],
                    ),
                ],
            )
        )

        assert energy == {
            "co2_grams_eq_per_kwh": 100,
            "schedules": [
                {"name": "default", "price_per_kwh": "246.8"},
                {
                    "name": "green",
                    "price_per_kwh": "123.4",
                    "periods": [
                        {"weekday": 1, "start_time": "23:00", "end_time": "00:00"}
                    ],
                },
            ],
        }

    @pytest.fixture
    def apps(self) -> AppsConfig:
        return AppsConfig(
            apps_hostname_templates=["{app_name}.apps.default.org.neu.ro"],
            app_proxy_url=URL("outputs-proxy.apps.default.org.neu.ro"),
        )

    def test_create_apps(self, factory: PayloadFactory, apps: AppsConfig) -> None:
        result = factory.create_apps(apps)
        assert result == {
            "apps_hostname_templates": ["{app_name}.apps.default.org.neu.ro"],
            "app_proxy_url": "outputs-proxy.apps.default.org.neu.ro",
        }



"""Platform config client."""

from importlib.metadata import version

from .client import ConfigClient, ConfigClientBase
from .entities import (
    AMDGPU,
    GPU,
    ACMEEnvironment,
    AddNodePoolRequest,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    AWSCloudProvider,
    AWSCredentials,
    AWSStorage,
    AzureCloudProvider,
    AzureCredentials,
    AzureReplicationType,
    AzureStorage,
    AzureStorageTier,
    BucketsConfig,
    CloudProvider,
    CloudProviderOptions,
    CloudProviderType,
    Cluster,
    ClusterLocationType,
    ClusterStatus,
    CredentialsConfig,
    DisksConfig,
    DNSConfig,
    DockerRegistryConfig,
    EFSPerformanceMode,
    EFSThroughputMode,
    EMCECSCredentials,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GoogleCloudProvider,
    GoogleFilestoreTier,
    GoogleStorage,
    GPUPreset,
    GrafanaCredentials,
    HelmRegistryConfig,
    IdleJobConfig,
    IngressConfig,
    IntelGPU,
    IntelGPUPreset,
    MetricsConfig,
    MinioCredentials,
    MonitoringConfig,
    NeuroAuthConfig,
    NodePool,
    NodePoolOptions,
    NodeRole,
    NotificationType,
    NvidiaGPU,
    NvidiaGPUPreset,
    OnPremCloudProvider,
    OpenStackCredentials,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchNodePoolResourcesRequest,
    PatchNodePoolSizeRequest,
    PatchOrchestratorConfigRequest,
    PrometheusCredentials,
    PutNodePoolRequest,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    SentryCredentials,
    Storage,
    StorageConfig,
    StorageInstance,
    TPUPreset,
    TPUResource,
    VCDCloudProvider,
    VCDCloudProviderOptions,
    VCDCredentials,
    VCDStorage,
    VolumeConfig,
)

__all__ = [
    "ConfigClient",
    "ConfigClientBase",
    "AMDGPU",
    "GPU",
    "ACMEEnvironment",
    "AddNodePoolRequest",
    "AMDGPUPreset",
    "AppsConfig",
    "ARecord",
    "AWSCloudProvider",
    "AWSCredentials",
    "AWSStorage",
    "AzureCloudProvider",
    "AzureCredentials",
    "AzureReplicationType",
    "AzureStorage",
    "AzureStorageTier",
    "BucketsConfig",
    "CloudProvider",
    "CloudProviderOptions",
    "CloudProviderType",
    "Cluster",
    "ClusterLocationType",
    "ClusterStatus",
    "CredentialsConfig",
    "DisksConfig",
    "DNSConfig",
    "DockerRegistryConfig",
    "EFSPerformanceMode",
    "EFSThroughputMode",
    "EMCECSCredentials",
    "EnergyConfig",
    "EnergySchedule",
    "EnergySchedulePeriod",
    "GoogleCloudProvider",
    "GoogleFilestoreTier",
    "GoogleStorage",
    "GPUPreset",
    "GrafanaCredentials",
    "HelmRegistryConfig",
    "IdleJobConfig",
    "IngressConfig",
    "IntelGPU",
    "IntelGPUPreset",
    "MetricsConfig",
    "MinioCredentials",
    "MonitoringConfig",
    "NeuroAuthConfig",
    "NodePool",
    "NodePoolOptions",
    "NodeRole",
    "NotificationType",
    "NvidiaGPU",
    "NvidiaGPUPreset",
    "OnPremCloudProvider",
    "OpenStackCredentials",
    "OrchestratorConfig",
    "PatchClusterRequest",
    "PatchNodePoolResourcesRequest",
    "PatchNodePoolSizeRequest",
    "PatchOrchestratorConfigRequest",
    "PrometheusCredentials",
    "PutNodePoolRequest",
    "RegistryConfig",
    "ResourcePoolType",
    "ResourcePreset",
    "Resources",
    "SecretsConfig",
    "SentryCredentials",
    "Storage",
    "StorageConfig",
    "StorageInstance",
    "TPUPreset",
    "TPUResource",
    "VCDCloudProvider",
    "VCDCloudProviderOptions",
    "VCDCredentials",
    "VCDStorage",
    "VolumeConfig",
]
__version__ = version(__name__)



from __future__ import annotations

import abc
import logging
from collections.abc import AsyncIterator, Mapping, Sequence
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from dataclasses import dataclass
from types import TracebackType
from typing import Any

import aiohttp
from aiohttp import ClientResponseError
from yarl import URL

from .entities import (
    AddNodePoolRequest,
    CloudProviderOptions,
    CloudProviderType,
    Cluster,
    NodePool,
    NotificationType,
    PatchClusterRequest,
    PatchNodePoolResourcesRequest,
    PatchNodePoolSizeRequest,
    PutNodePoolRequest,
    ResourcePreset,
)
from .factories import EntityFactory, PayloadFactory

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _Endpoints:
    clusters: str = "clusters"
    cloud_providers: str = "cloud_providers"

    def cloud_provider_options(self, type: CloudProviderType) -> str:
        return f"{self.cloud_providers}/{type.value}"

    def cluster(self, cluster_name: str) -> str:
        return f"{self.clusters}/{cluster_name}"

    def node_pools(self, cluster_name: str) -> str:
        return f"{self.cluster(cluster_name)}/cloud_provider/node_pools"

    def node_pool(self, cluster_name: str, node_pool_name: str) -> str:
        return f"{self.node_pools(cluster_name)}/{node_pool_name}"

    def storages(self, cluster_name: str) -> str:
        return f"{self.cluster(cluster_name)}/cloud_provider/storages"

    def storage(self, cluster_name: str, storage_name: str) -> str:
        return f"{self.storages(cluster_name)}/{storage_name}"

    def notifications(self, cluster_name: str) -> str:
        return f"{self.cluster(cluster_name)}/notifications"

    def resource_presets(self, cluster_name: str) -> str:
        return f"{self.cluster(cluster_name)}/orchestrator/resource_presets"

    def resource_preset(self, cluster_name: str, preset_name: str) -> str:
        return f"{self.resource_presets(cluster_name)}/{preset_name}"


class ConfigClientBase:
    def __init__(self) -> None:
        self._endpoints = _Endpoints()
        self._entity_factory = EntityFactory()
        self._payload_factory = PayloadFactory()

    @abc.abstractmethod
    def _request(
        self,
        method: str,
        path: str,
        json: dict[str, Any] | None = None,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> AbstractAsyncContextManager[aiohttp.ClientResponse]:
        pass

    def _create_headers(self, *, token: str | None = None) -> dict[str, str]:
        result = {}
        if token:
            result["Authorization"] = f"Bearer {token}"
        return result

    async def list_cloud_provider_options(
        self, *, token: str | None = None
    ) -> list[CloudProviderOptions]:
        path = self._endpoints.cloud_providers
        async with self._request(
            "GET", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            result: list[CloudProviderOptions] = []
            for k, v in resp_payload.items():
                result.append(
                    self._entity_factory.create_cloud_provider_options(
                        CloudProviderType(k), v
                    )
                )
            return result

    async def get_cloud_provider_options(
        self, type: CloudProviderType, *, token: str | None = None
    ) -> CloudProviderOptions:
        path = self._endpoints.cloud_provider_options(type)
        async with self._request(
            "GET", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cloud_provider_options(
                type, resp_payload
            )

    async def list_clusters(self, *, token: str | None = None) -> Sequence[Cluster]:
        async with self._request(
            "GET", self._endpoints.clusters, headers=self._create_headers(token=token)
        ) as response:
            payload = await response.json()
            return [self._entity_factory.create_cluster(p) for p in payload]

    async def get_cluster(self, name: str, *, token: str | None = None) -> Cluster:
        async with self._request(
            "GET",
            self._endpoints.cluster(name),
            headers=self._create_headers(token=token),
        ) as response:
            payload = await response.json()
            return self._entity_factory.create_cluster(payload)

    async def create_blank_cluster(
        self,
        name: str,
        service_token: str,
        *,
        ignore_existing: bool = False,
        token: str | None = None,
    ) -> Cluster:
        payload = {"name": name, "token": service_token}
        try:
            async with self._request(
                "POST",
                self._endpoints.clusters,
                headers=self._create_headers(token=token),
                json=payload,
            ) as resp:
                resp_payload = await resp.json()
                return self._entity_factory.create_cluster(resp_payload)
        except ClientResponseError as e:
            is_existing = e.status == 400 and "already exists" in e.message
            if not ignore_existing or is_existing:
                raise
        return await self.get_cluster(name)

    async def patch_cluster(
        self, name: str, request: PatchClusterRequest, *, token: str | None = None
    ) -> Cluster:
        path = self._endpoints.cluster(name)
        payload = self._payload_factory.create_patch_cluster_request(request)
        async with self._request(
            "PATCH", path, headers=self._create_headers(token=token), json=payload
        ) as resp:
            resp_payload = await resp.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def delete_cluster(self, name: str, *, token: str | None = None) -> None:
        async with self._request(
            "DELETE",
            self._endpoints.cluster(name),
            headers=self._create_headers(token=token),
        ):
            pass

    async def add_storage(
        self,
        cluster_name: str,
        storage_name: str,
        size: int | None = None,
        *,
        start_deployment: bool = True,
        ignore_existing: bool = False,
        token: str | None = None,
    ) -> Cluster:
        try:
            path = self._endpoints.storages(cluster_name)
            payload: dict[str, Any] = {"name": storage_name}
            if size is not None:
                payload["size"] = size
            async with self._request(
                "POST",
                path,
                params={"start_deployment": str(start_deployment).lower()},
                headers=self._create_headers(token=token),
                json=payload,
            ) as response:
                resp_payload = await response.json()
                return self._entity_factory.create_cluster(resp_payload)
        except ClientResponseError as e:
            if not ignore_existing or e.status != 409:
                raise
        return await self.get_cluster(cluster_name)

    async def patch_storage(
        self,
        cluster_name: str,
        storage_name: str | None,
        ready: bool | None = None,
        *,
        ignore_not_found: bool = False,
        token: str | None = None,
    ) -> Cluster:
        try:
            if storage_name:
                path = self._endpoints.storage(cluster_name, storage_name)
            else:
                path = self._endpoints.storage(cluster_name, "default/entry")
            payload: dict[str, Any] = {}
            if ready is not None:
                payload["ready"] = ready
            async with self._request(
                "PATCH", path, headers=self._create_headers(token=token), json=payload
            ) as response:
                resp_payload = await response.json()
                return self._entity_factory.create_cluster(resp_payload)
        except ClientResponseError as e:
            if not ignore_not_found or e.status != 404:
                raise
        return await self.get_cluster(cluster_name)

    async def remove_storage(
        self,
        cluster_name: str,
        storage_name: str,
        *,
        start_deployment: bool = True,
        ignore_not_found: bool = False,
        token: str | None = None,
    ) -> Cluster:
        try:
            path = self._endpoints.storage(cluster_name, storage_name)
            async with self._request(
                "DELETE",
                path,
                params={"start_deployment": str(start_deployment).lower()},
                headers=self._create_headers(token=token),
            ) as response:
                resp_payload = await response.json()
                return self._entity_factory.create_cluster(resp_payload)
        except ClientResponseError as e:
            if not ignore_not_found or e.status != 404:
                raise
        return await self.get_cluster(cluster_name)

    async def get_node_pool(
        self, cluster_name: str, node_pool_name: str, *, token: str | None = None
    ) -> NodePool:
        path = self._endpoints.node_pool(cluster_name, node_pool_name)
        async with self._request(
            "GET", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_node_pool(resp_payload)

    async def list_node_pools(
        self, cluster_name: str, *, token: str | None = None
    ) -> list[NodePool]:
        path = self._endpoints.node_pools(cluster_name)
        async with self._request(
            "GET", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            return [self._entity_factory.create_node_pool(n) for n in resp_payload]

    async def add_node_pool(
        self,
        cluster_name: str,
        node_pool: AddNodePoolRequest,
        *,
        start_deployment: bool = True,
        token: str | None = None,
    ) -> Cluster:
        """Add new node pool to the existing cluster.
        Cloud provider should be already set up.

        Make sure you use one of the available node pool templates by providing
        its machine type, if the cluster is deployed in public cloud
        (AWS / GCP / Azure / VCD).

        Args:
            cluster_name (str): Name of the cluster within the platform.
            node_pool (NodePool): Node pool instance.
                For templates, you could use template.to_node_pool() method
            start_deployment (bool, optional): Start applying changes. Defaults to True.

        Returns:
            Cluster: Cluster instance with applied changes
        """
        path = self._endpoints.node_pools(cluster_name)
        payload = self._payload_factory.create_add_node_pool_request(node_pool)
        async with self._request(
            "POST",
            path,
            params={"start_deployment": str(start_deployment).lower()},
            headers=self._create_headers(token=token),
            json=payload,
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def put_node_pool(
        self,
        cluster_name: str,
        node_pool: PutNodePoolRequest,
        *,
        start_deployment: bool = True,
        token: str | None = None,
    ) -> Cluster:
        path = self._endpoints.node_pool(cluster_name, node_pool.name)
        payload = self._payload_factory.create_add_node_pool_request(node_pool)
        async with self._request(
            "PUT",
            path,
            params={"start_deployment": str(start_deployment).lower()},
            headers=self._create_headers(token=token),
            json=payload,
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def patch_node_pool(
        self,
        cluster_name: str,
        node_pool_name: str,
        request: PatchNodePoolSizeRequest | PatchNodePoolResourcesRequest,
        *,
        start_deployment: bool = True,
        token: str | None = None,
    ) -> Cluster:
        path = self._endpoints.node_pool(cluster_name, node_pool_name)
        payload = self._payload_factory.create_patch_node_pool_request(request)
        async with self._request(
            "PATCH",
            path,
            params={"start_deployment": str(start_deployment).lower()},
            headers=self._create_headers(token=token),
            json=payload,
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def delete_node_pool(
        self,
        cluster_name: str,
        node_pool_name: str,
        *,
        start_deployment: bool = True,
        token: str | None = None,
    ) -> Cluster:
        path = self._endpoints.node_pool(cluster_name, node_pool_name)
        async with self._request(
            "DELETE",
            path,
            params={"start_deployment": str(start_deployment).lower()},
            headers=self._create_headers(token=token),
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def notify(
        self,
        cluster_name: str,
        notification_type: NotificationType,
        message: str | None = None,
        *,
        token: str | None = None,
    ) -> None:
        path = self._endpoints.notifications(cluster_name)
        payload = {"notification_type": notification_type.value}
        if message:
            payload["message"] = message
        async with self._request(
            "POST", path, headers=self._create_headers(token=token), json=payload
        ):
            pass

    async def list_resource_presets(
        self, cluster_name: str, *, token: str | None = None
    ) -> list[ResourcePreset]:
        path = self._endpoints.resource_presets(cluster_name)
        async with self._request(
            "GET", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            return [
                self._entity_factory.create_resource_preset(p) for p in resp_payload
            ]

    async def get_resource_preset(
        self, cluster_name: str, preset_name: str, *, token: str | None = None
    ) -> ResourcePreset:
        path = self._endpoints.resource_preset(cluster_name, preset_name)
        async with self._request(
            "GET", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_resource_preset(resp_payload)

    async def add_resource_preset(
        self, cluster_name: str, preset: ResourcePreset, *, token: str | None = None
    ) -> Cluster:
        path = self._endpoints.resource_presets(cluster_name)
        payload = self._payload_factory.create_resource_preset(preset)
        async with self._request(
            "POST", path, headers=self._create_headers(token=token), json=payload
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def put_resource_preset(
        self, cluster_name: str, preset: ResourcePreset, *, token: str | None = None
    ) -> Cluster:
        path = self._endpoints.resource_preset(cluster_name, preset.name)
        payload = self._payload_factory.create_resource_preset(preset)
        async with self._request(
            "PUT", path, headers=self._create_headers(token=token), json=payload
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)

    async def delete_resource_preset(
        self, cluster_name: str, preset_name: str, *, token: str | None = None
    ) -> Cluster:
        path = self._endpoints.resource_preset(cluster_name, preset_name)
        async with self._request(
            "DELETE", path, headers=self._create_headers(token=token)
        ) as response:
            resp_payload = await response.json()
            return self._entity_factory.create_cluster(resp_payload)


class ConfigClient(ConfigClientBase):
    def __init__(
        self,
        url: URL,
        token: str | None = None,
        timeout: aiohttp.ClientTimeout = aiohttp.client.DEFAULT_TIMEOUT,
        trace_configs: Sequence[aiohttp.TraceConfig] = (),
    ):
        super().__init__()

        self._base_url = url / "api/v1"
        self._token = token
        self._timeout = timeout
        self._trace_configs = trace_configs
        self._client: aiohttp.ClientSession | None = None

    async def __aenter__(self) -> ConfigClient:
        self._client = await self._create_http_client()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.aclose()

    async def aclose(self) -> None:
        assert self._client
        await self._client.close()

    async def _create_http_client(self) -> aiohttp.ClientSession:
        client = aiohttp.ClientSession(
            headers=self._create_default_headers(),
            timeout=self._timeout,
            trace_configs=list(self._trace_configs),
        )
        return await client.__aenter__()

    def _create_default_headers(self) -> dict[str, str]:
        result = {}
        if self._token:
            result["Authorization"] = f"Bearer {self._token}"
        return result

    @asynccontextmanager
    async def _request(
        self,
        method: str,
        path: str,
        json: dict[str, Any] | None = None,
        params: Mapping[str, str] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        assert self._client
        assert self._base_url
        url = self._base_url / path
        if params:
            url = url.with_query(params)

        async with self._client.request(
            method, url, json=json, headers=headers
        ) as response:
            response.raise_for_status()
            yield response



from __future__ import annotations

import abc
import enum
import sys
from collections.abc import Sequence
from dataclasses import dataclass, field
from datetime import datetime, time, tzinfo
from decimal import Decimal

from yarl import URL

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    # why not backports.zoneinfo: https://github.com/pganssle/zoneinfo/issues/125
    from backports.zoneinfo._zoneinfo import ZoneInfo


class NotificationType(str, enum.Enum):
    SUCCESS = "success"
    ERROR = "error"
    CLUSTER_UPDATING = "cluster_updating"
    CLUSTER_UPDATE_SUCCEEDED = "cluster_update_succeeded"
    CLUSTER_UPDATE_FAILED = "cluster_update_failed"


class CloudProviderType(str, enum.Enum):
    AWS = "aws"
    GCP = "gcp"
    AZURE = "azure"
    ON_PREM = "on_prem"
    VCD_MTS = "vcd_mts"
    VCD_SELECTEL = "vcd_selectel"

    @property
    def is_vcd(self) -> bool:
        return self.startswith("vcd_")


@dataclass(frozen=True)
class CloudProviderOptions:
    type: CloudProviderType
    node_pools: list[NodePoolOptions]


@dataclass(frozen=True)
class VCDCloudProviderOptions(CloudProviderOptions):
    url: URL | None = None
    organization: str | None = None
    edge_name_template: str | None = None
    edge_external_network_name: str | None = None
    catalog_name: str | None = None
    storage_profile_names: list[str] | None = None


@dataclass(frozen=True)
class NodePoolOptions:
    machine_type: str
    cpu: float
    memory: int
    available_cpu: float | None = None
    available_memory: int | None = None
    nvidia_gpu: int | None = None
    nvidia_gpu_model: str | None = None


class NodeRole(str, enum.Enum):
    KUBERNETES = "kubernetes"
    PLATFORM = "platform"
    PLATFORM_JOB = "platform_job"


@dataclass(frozen=True)
class NodePool:
    name: str

    cpu: float
    available_cpu: float
    memory: int
    available_memory: int
    disk_size: int
    available_disk_size: int

    role: NodeRole = NodeRole.PLATFORM_JOB

    min_size: int = 0
    max_size: int = 1
    idle_size: int | None = None

    machine_type: str | None = None

    disk_type: str | None = None

    gpu: int | None = None  # Deprecated. Use nvidia_gpu instead
    gpu_model: str | None = None  # Deprecated. Use nvidia_gpu_model instead
    nvidia_gpu: int | None = None
    nvidia_gpu_model: str | None = None
    amd_gpu: int | None = None
    amd_gpu_model: str | None = None
    intel_gpu: int | None = None
    intel_gpu_model: str | None = None

    price: Decimal | None = None
    currency: str | None = None

    is_preemptible: bool | None = None

    zones: tuple[str, ...] | None = None

    cpu_min_watts: float = 0.0
    cpu_max_watts: float = 0.0


@dataclass(frozen=True)
class AddNodePoolRequest:
    name: str

    min_size: int
    max_size: int
    idle_size: int | None = None

    role: NodeRole = NodeRole.PLATFORM_JOB

    machine_type: str | None = None

    cpu: float | None = None
    available_cpu: float | None = None
    memory: int | None = None
    available_memory: int | None = None
    disk_type: str | None = None
    disk_size: int | None = None
    available_disk_size: int | None = None

    nvidia_gpu: int | None = None
    nvidia_gpu_model: str | None = None
    amd_gpu: int | None = None
    amd_gpu_model: str | None = None
    intel_gpu: int | None = None
    intel_gpu_model: str | None = None

    price: Decimal | None = None
    currency: str | None = None

    is_preemptible: bool | None = None

    zones: tuple[str, ...] | None = None

    cpu_min_watts: float | None = None
    cpu_max_watts: float | None = None


PutNodePoolRequest = AddNodePoolRequest


@dataclass(frozen=True)
class PatchNodePoolSizeRequest:
    min_size: int | None = None
    max_size: int | None = None
    idle_size: int | None = None


@dataclass(frozen=True)
class PatchNodePoolResourcesRequest:
    cpu: float
    available_cpu: float
    memory: int
    available_memory: int
    disk_size: int
    available_disk_size: int

    nvidia_gpu: int | None = None
    nvidia_gpu_model: str | None = None
    amd_gpu: int | None = None
    amd_gpu_model: str | None = None
    intel_gpu: int | None = None
    intel_gpu_model: str | None = None

    machine_type: str | None = None

    min_size: int | None = None
    max_size: int | None = None


@dataclass(frozen=True)
class StorageInstance:
    name: str
    size: int | None = None
    ready: bool = False


@dataclass(frozen=True)
class Storage:
    instances: Sequence[StorageInstance]


@dataclass(frozen=True)
class CloudProvider(abc.ABC):
    node_pools: Sequence[NodePool]
    storage: Storage | None

    @property
    @abc.abstractmethod
    def type(self) -> CloudProviderType:
        pass


@dataclass(frozen=True, repr=False)
class AWSCredentials:
    access_key_id: str
    secret_access_key: str = field(repr=False)


class EFSPerformanceMode(str, enum.Enum):
    GENERAL_PURPOSE = "generalPurpose"
    MAX_IO = "maxIO"


class EFSThroughputMode(str, enum.Enum):
    BURSTING = "bursting"
    PROVISIONED = "provisioned"


@dataclass(frozen=True)
class AWSStorage(Storage):
    description: str
    performance_mode: EFSPerformanceMode
    throughput_mode: EFSThroughputMode
    provisioned_throughput_mibps: int | None = None


@dataclass(frozen=True)
class AWSCloudProvider(CloudProvider):
    region: str
    zones: Sequence[str]
    credentials: AWSCredentials = field(repr=False)
    storage: AWSStorage | None
    vpc_id: str | None = None

    @property
    def type(self) -> CloudProviderType:
        return CloudProviderType.AWS


class ClusterLocationType(str, enum.Enum):
    ZONAL = "zonal"
    REGIONAL = "regional"


class GoogleFilestoreTier(str, enum.Enum):
    STANDARD = "STANDARD"
    PREMIUM = "PREMIUM"


@dataclass(frozen=True)
class GoogleStorage(Storage):
    description: str
    tier: GoogleFilestoreTier


@dataclass(frozen=True)
class GoogleCloudProvider(CloudProvider):
    region: str
    zones: Sequence[str]
    project: str
    credentials: dict[str, str] = field(repr=False)
    location_type: ClusterLocationType = ClusterLocationType.ZONAL
    tpu_enabled: bool = False

    @property
    def type(self) -> CloudProviderType:
        return CloudProviderType.GCP


@dataclass(frozen=True)
class AzureCredentials:
    subscription_id: str
    tenant_id: str
    client_id: str
    client_secret: str = field(repr=False)


class AzureStorageTier(str, enum.Enum):
    STANDARD = "Standard"
    PREMIUM = "Premium"


class AzureReplicationType(str, enum.Enum):
    LRS = "LRS"
    ZRS = "ZRS"


@dataclass(frozen=True)
class AzureStorage(Storage):
    description: str
    tier: AzureStorageTier
    replication_type: AzureReplicationType


@dataclass(frozen=True)
class AzureCloudProvider(CloudProvider):
    region: str
    resource_group: str
    credentials: AzureCredentials
    virtual_network_cidr: str | None = None

    @property
    def type(self) -> CloudProviderType:
        return CloudProviderType.AZURE


@dataclass(frozen=True)
class KubernetesCredentials:
    ca_data: str
    token: str | None = field(repr=False, default=None)
    client_key_data: str | None = field(repr=False, default=None)
    client_cert_data: str | None = field(repr=False, default=None)


@dataclass(frozen=True)
class OnPremCloudProvider(CloudProvider):
    kubernetes_url: URL | None = None
    credentials: KubernetesCredentials | None = None

    @property
    def type(self) -> CloudProviderType:
        return CloudProviderType.ON_PREM


@dataclass(frozen=True)
class VCDCredentials:
    user: str
    password: str = field(repr=False)
    ssh_password: str = field(repr=False)


@dataclass(frozen=True)
class VCDStorage(Storage):
    description: str
    profile_name: str
    size: int


@dataclass(frozen=True)
class VCDCloudProvider(CloudProvider):
    _type: CloudProviderType
    url: URL
    organization: str
    virtual_data_center: str
    edge_name: str
    edge_public_ip: str
    edge_external_network_name: str
    catalog_name: str
    credentials: VCDCredentials

    @property
    def type(self) -> CloudProviderType:
        return self._type


@dataclass(frozen=True)
class NeuroAuthConfig:
    url: URL
    token: str = field(repr=False)


@dataclass(frozen=True)
class HelmRegistryConfig:
    url: URL
    username: str | None = None
    password: str | None = field(repr=False, default=None)


@dataclass(frozen=True)
class DockerRegistryConfig:
    url: URL
    username: str | None = None
    password: str | None = field(repr=False, default=None)
    email: str | None = None


@dataclass(frozen=True)
class GrafanaCredentials:
    username: str
    password: str = field(repr=False)


@dataclass(frozen=True)
class PrometheusCredentials:
    username: str
    password: str = field(repr=False)


@dataclass(frozen=True)
class SentryCredentials:
    client_key_id: str
    public_dsn: URL
    sample_rate: float = 0.01


@dataclass(frozen=True)
class MinioCredentials:
    username: str
    password: str = field(repr=False)


@dataclass(frozen=True)
class EMCECSCredentials:
    """
    Credentials to EMC ECS (blob storage engine developed by vmware creators)
    """

    access_key_id: str
    secret_access_key: str = field(repr=False)
    s3_endpoint: URL
    management_endpoint: URL
    s3_assumable_role: str


@dataclass(frozen=True)
class OpenStackCredentials:
    account_id: str
    password: str = field(repr=False)
    endpoint: URL
    s3_endpoint: URL
    region_name: str


@dataclass(frozen=True)
class CredentialsConfig:
    neuro: NeuroAuthConfig
    neuro_helm: HelmRegistryConfig
    neuro_registry: DockerRegistryConfig
    grafana: GrafanaCredentials | None = None
    prometheus: PrometheusCredentials | None = None
    sentry: SentryCredentials | None = None
    docker_hub: DockerRegistryConfig | None = None
    minio: MinioCredentials | None = None
    emc_ecs: EMCECSCredentials | None = None
    open_stack: OpenStackCredentials | None = None


@dataclass(frozen=True)
class VolumeConfig:
    name: str
    size: int | None = None
    path: str | None = None
    credits_per_hour_per_gb: Decimal = Decimal(0)


@dataclass(frozen=True)
class StorageConfig:
    url: URL
    volumes: Sequence[VolumeConfig] = ()


@dataclass(frozen=True)
class RegistryConfig:
    url: URL

    @property
    def host(self) -> str:
        """Returns registry hostname with port (if specified)"""
        port = self.url.explicit_port
        suffix = f":{port}" if port is not None else ""
        return f"{self.url.host}{suffix}"


@dataclass(frozen=True)
class MonitoringConfig:
    url: URL


@dataclass(frozen=True)
class MetricsConfig:
    url: URL


@dataclass(frozen=True)
class SecretsConfig:
    url: URL


@dataclass(frozen=True)
class DisksConfig:
    url: URL
    storage_limit_per_user: int


@dataclass(frozen=True)
class BucketsConfig:
    url: URL
    disable_creation: bool = False


class ACMEEnvironment(str, enum.Enum):
    STAGING = "staging"
    PRODUCTION = "production"


@dataclass(frozen=True)
class IngressConfig:
    acme_environment: ACMEEnvironment
    default_cors_origins: Sequence[str] = ()
    additional_cors_origins: Sequence[str] = ()


@dataclass(frozen=True)
class GPUPreset:
    count: int
    model: str | None = None
    memory: int | None = None


@dataclass(frozen=True)
class NvidiaGPUPreset(GPUPreset):
    pass


@dataclass(frozen=True)
class AMDGPUPreset(GPUPreset):
    pass


@dataclass(frozen=True)
class IntelGPUPreset(GPUPreset):
    pass


@dataclass(frozen=True)
class TPUPreset:
    type: str
    software_version: str


@dataclass(frozen=True)
class ResourcePreset:
    name: str
    credits_per_hour: Decimal
    cpu: float
    memory: int
    nvidia_gpu: NvidiaGPUPreset | None = None
    amd_gpu: AMDGPUPreset | None = None
    intel_gpu: IntelGPUPreset | None = None
    tpu: TPUPreset | None = None
    scheduler_enabled: bool = False
    preemptible_node: bool = False
    is_external_job: bool = False
    resource_pool_names: Sequence[str] = ()
    available_resource_pool_names: Sequence[str] = ()
    capacity: int = 0


@dataclass(frozen=True)
class GPU:
    count: int
    model: str
    memory: int | None = None


@dataclass(frozen=True)
class NvidiaGPU(GPU):
    pass


@dataclass(frozen=True)
class AMDGPU(GPU):
    pass


@dataclass(frozen=True)
class IntelGPU(GPU):
    pass


@dataclass(frozen=True)
class TPUResource:
    ipv4_cidr_block: str
    types: Sequence[str] = ()
    software_versions: Sequence[str] = ()


@dataclass(frozen=True)
class ResourcePoolType:
    name: str
    min_size: int = 0
    max_size: int = 1
    idle_size: int = 0

    cpu: float = 1.0
    available_cpu: float = 1.0
    memory: int = 2**30  # 1gb
    available_memory: int = 2**30
    disk_size: int = 150 * 2**30  # 150gb
    available_disk_size: int = 150 * 2**30  # 150gb

    nvidia_gpu: NvidiaGPU | None = None
    amd_gpu: AMDGPU | None = None
    intel_gpu: IntelGPU | None = None
    tpu: TPUResource | None = None

    price: Decimal = Decimal()
    currency: str | None = None

    is_preemptible: bool = False

    cpu_min_watts: float = 0.0
    cpu_max_watts: float = 0.0


@dataclass(frozen=True)
class Resources:
    cpu: float
    memory: int
    nvidia_gpu: int = 0
    amd_gpu: int = 0
    intel_gpu: int = 0


@dataclass(frozen=True)
class IdleJobConfig:
    name: str
    count: int
    image: str
    resources: Resources
    command: list[str] = field(default_factory=list)
    args: list[str] = field(default_factory=list)
    image_pull_secret: str | None = None
    env: dict[str, str] = field(default_factory=dict)
    node_selector: dict[str, str] = field(default_factory=dict)


@dataclass
class OrchestratorConfig:
    job_hostname_template: str
    job_fallback_hostname: str
    job_schedule_timeout_s: float
    job_schedule_scale_up_timeout_s: float
    is_http_ingress_secure: bool = True
    resource_pool_types: Sequence[ResourcePoolType] = ()
    resource_presets: Sequence[ResourcePreset] = ()
    allow_privileged_mode: bool = False
    allow_job_priority: bool = False
    pre_pull_images: Sequence[str] = ()
    idle_jobs: Sequence[IdleJobConfig] = ()

    @property
    def allow_scheduler_enabled_job(self) -> bool:
        for preset in self.resource_presets:
            if preset.scheduler_enabled:
                return True
        return False

    @property
    def tpu_resources(self) -> Sequence[TPUResource]:
        return tuple(
            resource.tpu for resource in self.resource_pool_types if resource.tpu
        )

    @property
    def tpu_ipv4_cidr_block(self) -> str | None:
        tpus = self.tpu_resources
        if not tpus:
            return None
        return tpus[0].ipv4_cidr_block


@dataclass
class PatchOrchestratorConfigRequest:
    job_hostname_template: str | None = None
    job_fallback_hostname: str | None = None
    job_schedule_timeout_s: float | None = None
    job_schedule_scale_up_timeout_s: float | None = None
    is_http_ingress_secure: bool | None = None
    resource_pool_types: Sequence[ResourcePoolType] | None = None
    resource_presets: Sequence[ResourcePreset] | None = None
    allow_privileged_mode: bool | None = None
    allow_job_priority: bool | None = None
    pre_pull_images: Sequence[str] | None = None
    idle_jobs: Sequence[IdleJobConfig] | None = None


@dataclass
class ARecord:
    name: str
    ips: Sequence[str] = ()
    dns_name: str | None = None
    zone_id: str | None = None
    evaluate_target_health: bool = False


@dataclass
class DNSConfig:
    name: str
    a_records: Sequence[ARecord] = ()


class ClusterStatus(str, enum.Enum):
    BLANK = "blank"
    DEPLOYING = "deploying"
    DESTROYING = "destroying"
    TESTING = "testing"
    DEPLOYED = "deployed"
    DESTROYED = "destroyed"
    FAILED = "failed"


@dataclass(frozen=True)
class EnergySchedulePeriod:
    # ISO 8601 weekday number (1-7)
    weekday: int
    start_time: time
    end_time: time

    @classmethod
    def create_full_day(cls, *, weekday: int, timezone: tzinfo) -> EnergySchedulePeriod:
        return cls(
            weekday=weekday,
            start_time=time.min.replace(tzinfo=timezone),
            end_time=time.max.replace(tzinfo=timezone),
        )


DEFAULT_ENERGY_SCHEDULE_NAME = "default"


@dataclass(frozen=True)
class EnergySchedule:
    name: str
    periods: Sequence[EnergySchedulePeriod] = ()
    price_per_kwh: Decimal = Decimal("0")

    @classmethod
    def create_default(
        cls, *, timezone: tzinfo, name: str = DEFAULT_ENERGY_SCHEDULE_NAME
    ) -> EnergySchedule:
        return cls(
            name=name,
            periods=[
                EnergySchedulePeriod.create_full_day(weekday=weekday, timezone=timezone)
                for weekday in range(1, 8)
            ],
        )

    def check_time(self, current_time: datetime) -> bool:
        return any(self._is_time_within_period(current_time, p) for p in self.periods)

    def _is_time_within_period(
        self, time_: datetime, period: EnergySchedulePeriod
    ) -> bool:
        time_ = time_.astimezone(period.start_time.tzinfo)
        return (
            period.weekday == time_.isoweekday()
            and period.start_time <= time_.timetz() < period.end_time
        )


@dataclass(frozen=True)
class EnergyConfig:
    co2_grams_eq_per_kwh: float = 0
    schedules: Sequence[EnergySchedule] = ()

    def get_schedule(self, name: str) -> EnergySchedule:
        return (
            self._get_schedule(name)
            or self._get_schedule(DEFAULT_ENERGY_SCHEDULE_NAME)
            or EnergySchedule.create_default(timezone=ZoneInfo("UTC"))
        )

    def _get_schedule(self, name: str) -> EnergySchedule | None:
        for schedule in self.schedules:
            if schedule.name == name:
                return schedule
        return None

    @property
    def schedule_names(self) -> list[str]:
        return [schedule.name for schedule in self.schedules]


@dataclass(frozen=True)
class AppsConfig:
    apps_hostname_templates: list[str]
    app_proxy_url: URL


@dataclass(frozen=True)
class Cluster:
    name: str
    status: ClusterStatus
    created_at: datetime
    orchestrator: OrchestratorConfig
    storage: StorageConfig
    registry: RegistryConfig
    monitoring: MonitoringConfig
    secrets: SecretsConfig
    metrics: MetricsConfig
    dns: DNSConfig
    disks: DisksConfig
    buckets: BucketsConfig
    ingress: IngressConfig
    energy: EnergyConfig
    apps: AppsConfig
    location: str | None = None
    logo_url: URL | None = None
    timezone: tzinfo = ZoneInfo("UTC")
    credentials: CredentialsConfig | None = None
    platform_infra_image_tag: str | None = None
    cloud_provider: CloudProvider | None = None


@dataclass(frozen=True)
class PatchClusterRequest:
    location: str | None = None
    logo_url: URL | None = None
    credentials: CredentialsConfig | None = None
    storage: StorageConfig | None = None
    registry: RegistryConfig | None = None
    orchestrator: PatchOrchestratorConfigRequest | None = None
    monitoring: MonitoringConfig | None = None
    secrets: SecretsConfig | None = None
    metrics: MetricsConfig | None = None
    disks: DisksConfig | None = None
    buckets: BucketsConfig | None = None
    ingress: IngressConfig | None = None
    dns: DNSConfig | None = None
    timezone: ZoneInfo | None = None
    energy: EnergyConfig | None = None
    apps: AppsConfig | None = None
